from phat.dists import CarBenHybrid, Phat, PhatStack
from phat.bootstrap import two_tailed_hill_double_bootstrap
from phat.learn import PhatNet, PhatLoss, PhatMetric, DataSplit
from phat.tseries import Garchcaster

__all__ = [
    "CarBenHybrid", "Phat", "PhatStack", 
    "PhatNet", "PhatLoss", "PhatMetric",
    "two_tailed_hill_double_bootstrap", "DataSplit", "Garchcaster"
]


"""
Collection of helper functions
"""

import numpy as np
import pandas as pd
from functools import wraps
from typing import Union, Iterable, Callable, Sequence

import matplotlib.pyplot as plt

XTYPE = Union[float, Iterable[Sequence[float]]]

def cosh_squared(x,k):
    return .5*(np.cosh(2*k*x) + 1)

def sech_squared(x,k):
    return 1 / cosh_squared(x, k)

def arrayarize(val:Iterable):
    """
    Converts list-like objects to np.ndarray
    """
    list_types = (list, tuple, set, pd.Series)
    if isinstance(val, np.ndarray):
        pass
    elif isinstance(val, list_types):
        val = np.array(val)
    elif isinstance(val, (int, float)):
        val = np.array([val])
    else:
        text = 'You must provide an iterable of type: '
        text += ', '.join(list_types)
        raise ValueError(text)
    return val

def argsetter(kws:Union[Iterable, str]='x', flat=False):
    if isinstance(kws, str):
        kws = list(kws)

    def deco(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if args:
                for k, arg in zip(kws, args):
                    kwargs[k] = arg if flat else arrayarize(arg)
            for k in kws:
                if k not in kwargs or kwargs[k] is None:                
                    kwargs[k] = getattr(self, k) if flat else arrayarize(getattr(self, k))
            
            return func(self, **kwargs)
    
        return wrapper

    return deco

def dotweight(func) -> Callable:
    """
    Takes the weighted average of values in the array `stack`
    Weights provided by array `p`. average found via dot product.

    Used as a wrapper for class methods: `p` must be an attribute of the class instance

    Returns
    --------
    np.ndarray; max 3 dimensions, the weighted averages along one dimension of `stack`
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs) -> np.ndarray:
        stack = func(self, *args, **kwargs)
        if stack.ndim == 3:
            p = np.tile(self.p, (stack.shape[0],stack.shape[1],1))
            return np.multiply(p, stack).sum(axis=2)
        else:
            return (self.p @ stack)

    return wrapper

def stacker(arr):
    """
    arr is numpy array
    """
    arr = arrayarize(arr)
    if arr.ndim == 1:
        return np.vstack
    elif arr.ndim == 2:
        return np.dstack
    else:
        raise ValueError(f'`arr` has {arr.ndim}')

class PriceSim:
    """
    Handler for generating price time-series from a set of return time-series
    """
    def __init__(self, p0:float, rets=None, periods:int=0, n:int=0):
        self.p0 = p0
        self.rets = rets
        self.n = n
        self.periods = periods
    
    @argsetter(['p0', 'rets', 'periods'])
    def sim(self, p0=None, rets=None, periods=None, show_chart:bool=False, *args, **kwargs):
        S = p0*rets.cumprod()
        if show_chart:
            return rets, S, self.sim_chart(rets, S, periods, *args, **kwargs)
        else:
            return rets, S

    @argsetter(['p0', 'rets'], flat=True)
    def sims(self, 
        p0:float=None, 
        rets=None, 
        show_chart:bool=False, 
        *args, **kwargs
    ):
        S = p0*rets.cumprod(axis=1)
        
        if show_chart:
            return rets, S, self.sims_chart(S, *args, **kwargs)
        else:
            return rets, S
        
    def sim_chart(self, rets, S, periods, axes:Iterable=None, title=''):
        if axes is None:
            fig, axes = plt.subplots(1,2,figsize=(14,5))
        elif len(axes) != 2:
            raise ValueError('Chart requires two subplots')
        
        ax1, ax2 = axes
        
        x = np.arange(periods)
        ax1.plot(x, rets - 1)
        ax2.plot(x, S)

        ax1.set_title('Daily Returns')
        ax2.set_title('Asset Price')

        if not title:
            title = 'Phat Price Simulation'
            
        plt.suptitle(title, y=1)
        
        return axes
    
    def sims_chart(self, S, ax=None, title='', *args, **kwargs):
        if ax is None:
            fig, ax = plt.subplots(1,1,figsize=(10,6))

        if 'bins' in kwargs:
            if isinstance(kwargs['bins'], (float, int)):
                bins = np.linspace(0, S[:, -1].max(), kwargs['bins'])
            else:
                bins = kwargs['bins']
            kwargs.pop('bins')
        else:
            bins = np.linspace(0, S[:, -1].max(), 250)
                                
        counts, bins, _ = ax.hist(S[:, -1], bins=bins, density=True, *args, **kwargs)

        if not title:
            title = 'Distribution of Ending Share Prices'
        
        ax.set_title(title)

        return ax, bins



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

import tensorflow as tf

logger = tf.get_logger()

def hill_est(data, return_details=False):
    """
    Data is sorted in reverse order (i.e. kth largest first)
    
    Adapted from: https://github.com/ivanvoitalov/tail-estimation
    """
    logs = tf.math.log(data)
    k = tf.range(1., data.shape[0], dtype=tf.float64)
    M1 = (tf.constant(1., dtype=tf.float64)/k)*tf.math.cumsum(logs[:-1]) - logs[1:]
    return M1, logs, k

def second_moment(data, logs1, k=None):
    if k is None:
        k = tf.range(1., tf.shape(data)[0])
    
    logs2 = tf.math.log(data)**tf.constant(2., dtype=tf.float64)
    term1 = (1/tf.cast(k, dtype=np.float64))*tf.math.cumsum(logs2[:-1])
    term2 = (tf.constant(2., dtype=tf.float64)*logs1[1:]/tf.cast(k, dtype=np.float64))*tf.math.cumsum(logs1[:-1])
    M2 = term1 - term2 + logs2[1:]
    
    return M2
    
def moments_estimator(data):
    """
    Function to calculate second moments array
    given an ordered data sequence. 
    Decreasing ordering is required.
    Args:
        ordered_data: numpy array of ordered data for which
                      the 1st (Hill estimator) and 2nd moments 
                      are calculated.
    Returns:
        M2: numpy array of 2nd moments corresponding to all 
            possible order statistics of the dataset.
    """

    M1, logs1, k = hill_est(data, return_details=True)
    M2 = second_moment(data, logs1, k)
    return M1, M2

def amse(M1, M2):
    """
    Asymptotic Mean Squared Error derived from first and second moments
    
    M1 is adjusted to account for different multiplicative constants in the two moments
    """
    return (M2 - 2*M1**2)**2

def bootstrap(data, n, r):
    errs = tf.zeros((r, n-1), dtype=tf.float64)
    valid = tf.zeros((r, n-1), dtype=tf.float64)
    for i in range(r):
        sample = tf.numpy_function(np.random.choice, [data, n, tf.constant(True)], tf.float64)
        # sample = np.random.choice(data, n, replace=True)
        sample = tf.sort(sample, direction='DESCENDING')
        M1, M2 = moments_estimator(sample)
        
        err = M2 - tf.constant(2, dtype=tf.float64)*M1**tf.constant(2, dtype=tf.float64)
        val = tf.cast(~tf.math.is_nan(errs[i]), dtype=tf.float64)

        idxs_to_update = tf.stack((tf.repeat(i, n-1), tf.range(n-1)), axis=1)
        tf.tensor_scatter_nd_update(errs, idxs_to_update, err)
        tf.tensor_scatter_nd_update(valid, idxs_to_update, val)

    return tf.math.reduce_sum(errs, axis=0)**tf.constant(2, dtype=tf.float64) / tf.math.reduce_sum(valid, axis=0)

def k_finder(data, idx, n, r, t):
    avg_amses = bootstrap(data[idx:], n, r)
    return tf.cast(tf.math.argmin(avg_amses), dtype=tf.int32) + tf.constant(1, dtype=tf.int32) + idx

def rho_dani(n1, k1):
    return (tf.math.log(k1)/(2.*np.log(n1) - np.log(k1)))\
          **(2.*(np.log(n1) - np.log(k1))/(np.log(n1)))

def rho_qi(n1, k1):
    return (tf.constant(1, dtype=tf.float64) - (tf.constant(2, dtype=tf.float64)*(tf.math.log(k1) - tf.math.log(n1))/(tf.math.log(k1))))**(tf.math.log(k1)/tf.math.log(n1) - tf.constant(1, dtype=tf.float64))

def hill_dbl_bs(data, t=.5, r=500, rho_type='qi'):
    """
    Different research provides different \rho values. we use
    rho_dani constant is given in Corollary 7 of Danielsson's paper
    https://www.sciencedirect.com/science/article/pii/S0047259X00919031
    rho_qi 
    https://link.springer.com/article/10.1007%2Fs10687-007-0049-8
    
    # eps_stop = 1
    # eps = 0.5*(1+np.log(int(t*n))/np.log(n)) # ???????
    # max_index1 = (np.abs(np.linspace(1/n1, 1.0, n1) - eps_stop)).argmin()
    """
    t = tf.cast(t, dtype=tf.float64)
    n = tf.cast(tf.size(data), dtype=tf.float64)
    xi, _, __ = hill_est(data)
    if logger.level <= 10:
        tf.print('Init Hill Estimates: ', xi[0])
    min1 = min2 = tf.constant(1, dtype=tf.int32)
    n1 = tf.cast(tf.multiply(tf.math.sqrt(t), n), dtype=tf.int32)
    n2 = tf.cast(tf.multiply(t, n), dtype=tf.int32)
    k1 = k2 = tf.constant(0, dtype=np.int32)
    
    for i in tf.range(tf.size(data)):
        k1 = k_finder(data, min1, n1, r, t)
        k2 = k_finder(data, min2, n2, r, t)
     
        if k2 > k1:
            min1 += tf.cast(tf.multiply(tf.constant(0.005, dtype=tf.float64), tf.cast(n1, dtype=tf.float64)), dtype=tf.int32)
            min2 += tf.cast(tf.multiply(tf.constant(0.005, dtype=tf.float64), tf.cast(n1, dtype=tf.float64)), dtype=tf.int32)
        else:
            break

    if rho_type == 'dani':
        rho = rho_dani(tf.cast(n1, dtype=tf.float64), tf.cast(k1, dtype=tf.float64)) 
    elif rho_type =='qi':
        rho = rho_qi(tf.cast(n1, dtype=tf.float64), tf.cast(k1, dtype=tf.float64)) 
    else:
        raise ValueError("`rho_type` must be either 'dani' or 'qi'")
    
    tf.print('Ns', n, n1, n2)
    tf.print('k1', k1, 'k2', k2)
    k_star = tf.cast(tf.math.round(rho*tf.cast(k1, dtype=tf.float64)**tf.constant(2, dtype=tf.float64)/tf.cast(k2, dtype=tf.float64)), dtype=tf.int32)

    if k_star >= tf.cast(n, dtype=tf.int32):
        tf.print('k1', k1, 'k2', k2)
        tf.print("WARNING: estimated threshold k is larger than the size of data")
        k_star = tf.cast(n, dtype=tf.int32)-tf.constant(1, dtype=tf.int32)
    elif k_star == tf.constant(0, dtype=tf.int32):
        k_star = tf.constant(2, dtype=tf.int32)
    else:
        k_star = k_star

    return xi[k_star-1], k_star

def conc_circles(alphas, ax=None, u50=False):
    if ax is None:
        fig, ax = plt.subplots()
    
    n_u10 = (np.abs(alphas[:,2]) < .1).sum()
    u10_per = n_u10 / alphas.shape[0]
    n_u25 = (np.abs(alphas[:,2]) < .25).sum()
    u25_per = n_u25 / alphas.shape[0]
    n_u50 = (np.abs(alphas[:,2]) < .5).sum()
    u50_per = n_u50 / alphas.shape[0]

    ax.add_patch(mpatches.Circle(
        (.5,0), 1, fc='grey', ec='grey', 
        ls='--', lw=1, alpha=.5, zorder=3
    ))
    ax.add_patch(mpatches.Circle((.5,0), u25_per, fc='C1', zorder=5))
    ax.add_patch(mpatches.Circle((.5,0), u10_per, fc='C0', zorder=6))

    ax.text(.5, .9, f'Total: {alphas.shape[0]:.0f}', ha='center')
    ax.text(.5, .75*u25_per, f'Error <25%: {n_u25:.0f}, {u25_per:.0%}', ha='center', zorder=5)
    ax.text(.5, .75*u10_per, f'Error <10%: {n_u10:.0f}, {u10_per:.0%}', ha='center', zorder=6)

    if u50:
        ax.add_patch(mpatches.Circle((.5,0), u50_per, fc='C2', zorder=4))
        ax.text(.5, .75*u50_per, f'Error <50%: {n_u50:.0f}, {u50_per:.0%}', ha='center', zorder=4)

    ax.axis('off')
    ax.set_title('Accuracy of Hill Double BootStrap')



"""
Tensorflow does not support scipy, therefore, all distributions in 
phat/dists.py must be replicated via tensorflow_probability 
"""

import numpy as np

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
import tensorflow_probability as tfp
from tensorflow_probability import distributions as tfd

from tensorflow_probability.python.internal import reparameterization, dtype_util, \
    tensor_util, parameter_properties
from tensorflow_probability.python.bijectors import softplus as softplus_bijector, identity as identity_bijector

### DEBUG UTILS ###
@tf.function
def find_nans(val):
    cond = tf.math.is_nan(val)
    idx = tf.where(cond)
    return idx

@tf.function
def find_zeros(val):
    cond = tf.math.equal(val, 0)
    idx = tf.where(cond)
    return idx

@tf.function
def if_not_empty(x, idx):
    if tf.not_equal(tf.size(idx), 0):
        return x[idx[0,0]]
    else:
        return tf.constant(0., dtype=np.float64)

# idx = find_nans(####)
# tf.print('loc', self.loc[0], self.scale[0], self.mean[0], self.std[0], self.shape[0])
# tf.print('idx', if_not_empty(x, idx), -if_not_empty(x, idx) + 10**-1, idx)

def gptf_prob(
    gptf,
    x,
    model
    ):
    loc = tf.convert_to_tensor(gptf.loc)
    scale = tf.convert_to_tensor(gptf.scale)
    concentration = tf.convert_to_tensor(gptf.concentration)
    conc_zero = tf.equal(concentration, 0)
    conc = tf.where(conc_zero, tf.constant(1, gptf.dtype), concentration)
    conc_neg = tf.math.less(conc, 0)
    neg_max = loc - (scale / conc)

    z = gptf._z(x, scale)
    base = tf.ones_like(z, gptf.dtype) + z*conc
    pow_ = -1 / conc - tf.ones_like(z, gptf.dtype)
    # tf.print('\n****RIGHT TAIL???', model.rtail)
    # tf.print('\nXXXXXXX', x)
    # tf.print('\nLOC', loc)
    # tf.print('\nSCALE', scale)
    # tf.print('\nZZZZZZZ', z)
    # tf.print('\nCONC', conc)
    # tf.print('\nBASE', base)
    # tf.print('POW_', pow_)
    
    base_pow_ = tf.math.pow(base, pow_) / gptf.scale

    # idx3 = find_nans(base_pow_)
    # tf.print('BASE_POW_', if_not_empty(base_pow_, idx3), idx3)
    # tf.print(base_pow_)
    
    # if tf.not_equal(tf.size(idx3), 0):
    #     tf.print('RTail?', model.rtail)
    #     tf.print('BASE_POW_', tf.gather_nd(base_pow_, idx3))
    #     tf.print (tf.shape(conc), tf.shape(idx3))
        
    #     tf.print('X', tf.shape(x), x[idx3[0,0]])
        
    #     tf.print('Loc', loc[idx3[0,0]], tf.unique(loc))
    #     tf.print(x<loc[0])
    #     tf.print('X<loc', tf.shape(x[x<loc[0]]))
    #     tf.print('Scale', tf.shape(gptf.scale), tf.unique(gptf.scale))
    #     tf.print('Z', tf.shape(z), tf.unique(z[idx3[0,0]]))
    #     tf.print('CONC_', conc[idx3[0,0]])
    #     tf.print('BASE', tf.shape(base[idx3[0,0]]), tf.unique(base[idx3[0,0]]))
        
    #     inval_conc = conc[idx3[0,0]]
    #     inval_z = z[idx3[0,0]][0]
    #     inval_scale = gptf.scale[0]
    #     inval_base = 1 + inval_z*inval_conc
    #     inval_pow = -1 / inval_conc - 1
        
    #     tf.print('inval conc', inval_conc, 'inval base', inval_base, 'inval pow', inval_pow)
    #     inval_base_pow = tf.math.pow(inval_base, inval_pow) / inval_scale
    #     tf.print('inval base pow, PRE SCALE', tf.math.pow(inval_base, inval_pow))
    #     tf.print('inval base pow', inval_base_pow)

    # def base_pow_(x, scale, conc, gptf):
        # return base_pow_

    # valid_x = tf.greater_equal(x, loc[0])
    # w_valid_x = tf.where(valid_x, calc_prob(x), tf.constant(0, gptf.dtype))

    w_negconc = tf.where((x>=loc) & (x<=neg_max), base_pow_, 0+10**-10)
    w_posconc = tf.where(x >= loc, base_pow_, 0+10**-10)
    
    return tf.where(
        conc_zero, z, tf.where(conc_neg, w_negconc, w_posconc)
    )

class CarbenBase4TF(tfd.Distribution):    
    def __init__(self, mean, std, shape,
        validate_args=False, 
        allow_nan_stats=True,
        name='CarbenBase'
        ):
        parameters = dict(locals())
        with tf.name_scope(name) as name:
            dtype = tf.float64
            self._mean = tensor_util.convert_nonref_to_tensor(mean, dtype=dtype, name='mean')
            self._std = tensor_util.convert_nonref_to_tensor(std, dtype=dtype, name='std')
            self._shape = tensor_util.convert_nonref_to_tensor(shape, dtype=dtype, name='shape')
            self._loc = self._calc_loc()
            self._scale = self._calc_scale()
            self._body = tfd.Normal(self._mean, self._std)
            super(CarbenBase4TF, self).__init__(
              dtype=dtype,
              reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
              validate_args=validate_args,
              allow_nan_stats=allow_nan_stats,
              parameters=parameters,
              name=name
            )

    @classmethod
    def _parameter_properties(cls, dtype, num_classes=None):
    # pylint: disable=g-long-lambda
        return dict(
            mean=parameter_properties.ParameterProperties(),
            std=parameter_properties.ParameterProperties(
                default_constraining_bijector_fn=(
                    lambda: softplus_bijector.Softplus(low=dtype_util.eps(dtype)))
            ),
            shape=parameter_properties.ParameterProperties()
        )
        # pylint: enable=g-long-lambda
        
    @property
    def mean(self):
        return self._mean
    
    @property
    def std(self):
        return self._std
    
    @property
    def shape(self):
        return self._shape

    @property
    def loc(self):
        return self._loc

    @property
    def scale(self):
        return self._scale

    @property
    def parameters(self):
        params = super().parameters
        params['loc'] = self.loc
        params['scale'] = self.scale
        return params

    def _z_for_W(self):
        """
        in Carreau (2008), z defined as:

            (1 + shape)**2 / 2*pi
        """
        num = (1 + self.shape)**2
        denom = 2*np.pi
        return  num / denom        

    def _W_z(self):
        """
        in Carreau (2008), W(z) can be calculated directly as:
            
            std**2 * (1 + shape)**2 / scale**2

        scale is not a free parameter, however, so we must calculate W(z) via
        the Lambert function on z, which is defined only in free parameters.
        """

        return tfp.math.lambertw(self._z_for_W())

    @property
    def gamma(self):
        val = tf.math.sqrt(self._W_z() / 2)
        return 1 + .5*(1 + tf.math.erf(val))
    
    def _calc_scale(self):
        num = self._std*(1 + self._shape)
        denom = tf.math.sqrt(self._W_z())
        return num / denom

    @property
    def body(self):
        if hasattr(self, '_body'):
            return self._body
        else:
            raise NotImplementedError

    @property
    def tail(self):
        if hasattr(self, '_tail'):
            return self._tail
        else:
            raise NotImplementedError

    def _log_prob(self, x):
        prob = self._prob(x)
        return tf.math.log(prob)

    def _event_shape_tensor(self):
        return tf.constant([], dtype=tf.int32)

    def _event_shape(self):
        return tf.TensorShape([])
    
    def _default_event_space_bijector(self):
        return identity_bijector.Identity(validate_args=self.validate_args)

class CarbenRight4TF(CarbenBase4TF):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tail = tfd.GeneralizedPareto(loc=self._loc, scale=self._scale, concentration=self._shape)
    
    def _calc_loc(self):
        return self.mean + self.std*tf.math.sqrt(self._W_z())

    def _prob(self, x):
        """
        Right tail only.
        where x>a,
            f(x) = f_t(x) / gamma
        """
        x_tail = x > self.loc
        # tf.print('is _prob: any tail????', x_tail)
        # tf.print('Rtail?', self.rtail, 'does this work?', tf.shape(x[x>self.loc[0]]))
        prob = tf.where(
            x_tail,
            gptf_prob(self.tail, tf.where(x_tail, x, self.loc), self),
            self.body.prob(x) + 10**-10,
        )
        return prob / self.gamma

class CarbenLeft4TF(CarbenBase4TF):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tail = tfd.GeneralizedPareto(loc=-self._loc, scale=self._scale, concentration=self._shape)
    
    def _calc_loc(self):
        return self.mean - self.std*tf.math.sqrt(self._W_z())

    def _prob(self, x):
        """
        Left tail only.
        where x>a,
            f(x) = f_t(x) / gamma
        """
        x_tail = x < self.loc
        # tf.print ('in _prob: right taill?????', self.rtail)
        # tf.print('in _prob: xxxxxx', x)
        # tf.print('in _prob: LOC', self.loc)
        # tf.print('is _prob: any tail????', x_tail)
        prob = tf.where(
            x < self.loc,
            gptf_prob(self.tail, tf.where(x_tail, -x, -self.loc), self),
            self.body.prob(x) + 10**-10,
        )
        return prob / self.gamma

class CarbenHybrid4TF:
    def __new__(cls, *args, **kwargs):
        args = list(args)
        shape_is_arg =  len(args) >= 2
        shape = args[2] if shape_is_arg else kwargs['shape']
        
        rtail_is_arg = len(args) >= 4
        if rtail_is_arg:
            rtail = args.pop(3)
        else:
            rtail = kwargs.pop('rtail') if 'rtail' in kwargs else True

        if isinstance(shape, float):
            if shape < 0 and not rtail:
                txt = 'If you provide a negative shape parameter,'
                txt += ' do not provide `rtail`'
                txt += ' as a left-tailed Carben is computed automatically.'
                raise ValueError(txt)             
            elif shape < 0:
                rtail = False
                if shape_is_arg:
                    args[2] = -shape
                else:
                    kwargs['shape'] = -shape
        
        if rtail:
            kwargs['name'] = 'CarbenRight'
            obj = CarbenRight4TF(*args, **kwargs)
            obj.rtail = True
        else:
            kwargs['name'] = 'CarbenLeft'            
            obj = CarbenLeft4TF(*args, **kwargs)
            obj.rtail = False
            
        return obj

class PhatMixture(tfd.Mixture):
    def __repr__(self):
        prepend = '<Phat4TF (inherits from tfp.distributions.Mixture) '
        append = ' '.join(super().__repr__().split(' ')[2:])
        return prepend + append

    def __str__(self):
        return str(self.__repr__())

class Phat4TF:
    def __new__(cls, mean, std, shape_l, shape_r, mix=.5):
        m = tf.shape(mean)[0]
        mix = tf.constant([mix, 1-mix], dtype=tf.float64)
        p = tf.ones([m, 1], dtype=tf.float64) * mix
        
        c1 = CarbenHybrid4TF(mean, std, shape_l, rtail=False)
        c2 = CarbenHybrid4TF(mean, std, shape_r)
        
        obj = PhatMixture(
            cat=tfd.Categorical(probs=p),
            components=[c1,c2]
        )
        obj.left = obj.components[0]
        obj.right = obj.components[1]
        return obj




from typing import Union, Iterable
import numpy as np
import functools
from collections import namedtuple

import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.animation import FuncAnimation

from sklearn.model_selection import train_test_split
import tensorflow as tf

from phat.utils import arrayarize

def nnelu(input):
    """ Computes the Non-Negative Exponential Linear Unit
    """
    return tf.add(tf.constant(1, dtype=tf.float64), tf.nn.elu(input))

def asymmactiv(x):
    lam = tf.constant(1, dtype=tf.float64)
    term1 = tf.constant(1, dtype=tf.float64) + tf.math.multiply(lam, tf.math.exp(x))
    return tf.constant(1, dtype=tf.float64) - term1**(-1./lam)

def splitter(func):
    @functools.wraps(func)
    def wrap(self, data):
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data
        return func(self, x, y, sample_weight)
    
    return wrap

class GraphicMixin:
    def build_graph(self):
        inputs = tf.keras.Input(shape=(1))
        return tf.keras.Model(inputs=[inputs], outputs=self.call(inputs))

    def plot_model(self, show_shapes:bool=True, to_file:str='phatdn.png', *args, **kwargs):
        return tf.keras.utils.plot_model(
            self.build_graph(),
            to_file=to_file,
            show_shapes=show_shapes
        )

    def loss_progress(self, history=None, metric='', ax=None):
        history = self.history if history is None else history
        
        if ax is None:
            fig, ax = plt.subplots()

        C2 = plt.rcParams['axes.prop_cycle'].by_key()['color'][1]
        C2_rgba = np.ones(4)
        C2_rgba[:3] = mcolors.to_rgb(C2)
        C2_rgba[3] = 0
        dots = plt.scatter([0], [0], marker='o', color=C2_rgba, s=100)
        vloss = np.array(history.history['val_loss'])
        vloss = vloss[~np.isnan(vloss)]

        metric = metric if metric else 'val_mean'
        vmean = np.array(history.history[metric])
        vmean = vmean[~np.isnan(vloss)]

        mumin, mumax = np.array(vmean).min(), \
            np.array(vmean).max()
        # stdmin, stdmax = np.array(history.history['val_std']).min(), \
        #     np.array(history.history['val_std']).max()
        lossmin, lossmax = np.array(vloss).min(), \
            np.array(vloss).max()

        # ms = np.linspace(mumin,mumax)
        # stds = np.linspace(stdmin,stdmax)
        ax.axvline(vmean[-1], c='C4', ls='--', lw=2)

        ax.set_xlim((
            mumin*1.1 if mumin <=0 else mumin*.9, 
            mumax*1.1 if mumax >=0 else mumax*.9,
        ))    
        ax.set_ylim((
            lossmin*1.1 if lossmin <=0 else lossmin*.9, 
            lossmax*1.1 if lossmax >=0 else lossmax*.9,
        ))

        datetext = ax.text(
            0, 1.1, '', 
            size=16, alpha=1, fontweight='bold', 
            ha='left', va='top', transform=ax.transAxes
        )

        def snap(i):
            alphas = np.tanh(np.linspace(3,0,i+1))
            if alphas.size > 1:     
                alphas[1:] /= 8

            colors = np.zeros((i+1, 4))
            colors[:i+1, :3] = C2_rgba[:3]
            colors[:i+1, 3] = alphas[::-1]

            vals = np.vstack((
                vmean[:i+1], 
                vloss[:i+1]
            )).T
            dots.set_offsets(vals)
            dots.set_color(colors)
            datetext.set_text(f'Epoch {i}')

            return dots

        anime = FuncAnimation(fig, snap, blit=False, 
            frames=history.epoch, 
            interval=200, repeat=False
        )
        return anime

class DataSplit:
    #     https://www.baeldung.com/cs/data-normalization-before-after-splitting-set
    # Split validaiton set, THEN apply normalizaiton and then split test set
    XY = namedtuple('XY', ['x', 'y'])

    def __init__(self, 
        y, x:Iterable[float]=None, 
        test_size:float=.1, val_size:float=.1,
        batch_sizes:Union[int, Iterable]=32,
        preprocess:str=''
        ):
        train_batch, test_batch, val_batch = self._set_batch_sizes(batch_sizes)

        y = arrayarize(y)
        x = arrayarize(x) if x is not None else np.zeros(y.size)
        self.raw = self.XY(x.copy(), y.copy())
        self.test_size = test_size
        self.val_size = val_size

        x, xt, y, yt = train_test_split(x.reshape(-1,1), y.reshape(-1,1), test_size=val_size, random_state=42)
        
        if preprocess == 'minmax':
            y = self.rescale(y)
        elif preprocess == 'minmaxminus':
            y = 2*self.rescale(y) - 1
        elif preprocess == 'allpos':
            y = self.pos_shift(y)
        elif preprocess == 'tanh':
            y = np.tanh(y)

        x, xval, y, yval = train_test_split(x, y, test_size=test_size, random_state=42)
        
        self.train_raw = self.XY(x.copy(), y.copy())
        self.train = tf.data.Dataset.from_tensor_slices((x, y)).batch(train_batch)
        self.test_raw = self.XY(xt.copy(), yt.copy())
        self.test = tf.data.Dataset.from_tensor_slices((xt, yt)).batch(test_batch)
        self.val_raw = self.XY(xval.copy(), yval.copy())
        self.val = tf.data.Dataset.from_tensor_slices((xval, yval)).batch(val_batch)

    def _set_batch_sizes(self, batch_sizes):
        if isinstance(batch_sizes, (float, int)):
            return list([batch_sizes])*3
        elif isinstance(batch_sizes, (list, tuple)):
            if len(batch_sizes) == 3:
                return batch_sizes
            elif len(batch_sizes) == 2:
                batch_sizes = list(batch_sizes)
                batch_sizes.append(32)
                return batch_sizes
            else:
                raise ValueError('`batch_sizes` can only be length 2/3')
        else:
            raise ValueError('`batch_sizes` can be int, list, or dict')

    def rescale(self, y):
        num = y - y.min()
        denom = y.max() - y.min()
        return num / denom

    def descale(self, y_scaled):
        return (y_scaled*(self.raw.y.max() - self.raw.y.min())) + self.raw.y.min()

    def pos_shift(self, y):
        y_ = y - y.min() if y.min() < 0 else y
        return y_



import numpy as np
import pandas as pd
import itertools as it

import tensorflow as tf

from phat.learn.dists import Phat4TF
from phat.learn.utils import GraphicMixin, splitter

logger = tf.get_logger()

class PhatMetric(tf.keras.metrics.Metric):
    """
    Custom tensorflow metric that updates the Phat distribution parameters
    """
 
    param_names = ['mean', 'std', 'loc', 'scale', 'shape']
    ends = ['right', 'left']
    ALLOWED_NAMES = ['_'.join(tup) for tup in it.product(param_names, ends)]
    
    def __init__(self, name, **kwargs):
        if not name in self.ALLOWED_NAMES:
            raise ValueError(f"`name` must be one of {', '.join(self.ALLOWED_NAMES)}")
        super(PhatMetric, self).__init__(name=name, **kwargs)
        self.param = self.add_weight(name='param', initializer='zeros')
        
    def update_state(self, y, phat, sample_weight=None):
        p_name = self.name.split('_')
        param = getattr(getattr(phat, p_name[1]), p_name[0])
        self.param.assign(tf.reduce_mean(param))

    def result(self):
        return self.param

    def reset_state(self):
        self.param.assign(0)

class AMLSE(tf.keras.metrics.Metric):
    ALLOWED_NAMES = ['left', 'right']
    def __init__(self, shape, name, **kwargs):
        if not name in self.ALLOWED_NAMES:
            raise ValueError(f"`name` must be one of {', '.join(self.ALLOWED_NAMES)}")        
        super(AMLSE, self).__init__(name=f'amlse_{name}', **kwargs)
        self.shape = shape
        self.tail = self.name[5:]
        self.amlse = self.add_weight(name='amlse', initializer='zeros')

    @staticmethod
    def calc_amlse(shape_, shape):
        return tf.reduce_mean((tf.math.log(shape_) - tf.math.log(shape))**2)
        
    @staticmethod
    def shapes_for_tail(phat, tail, shape):
        shape_ = getattr(phat, tail).shape
        shape = tf.ones_like(shape_)*shape
        return shape_, shape
    
    @staticmethod
    def calc(phat, tail, shape):
        shape_, shape = AMLSE.shapes_for_tail(phat, tail, shape)
        return AMLSE.calc_amlse(shape_, shape)
    
    def update_state(self, y, phat, sample_weight=None):
        self.amlse.assign(AMLSE.calc(phat, self.tail, self.shape))
                          
    def result(self):
        return self.amlse

    def reset_state(self):
        self.amlse.assign(0)

class TwoTailedAMLSE(tf.keras.metrics.Metric):
    def __init__(self, shape_l, shape_r, name='two_tailed_amlse', **kwargs):
        super(TwoTailedAMLSE, self).__init__(name=name, **kwargs)
        self.shape_l = shape_l
        self.shape_r = shape_r       
        self.two_tailed = self.add_weight(name='two_tailed_amlse', initializer='zeros')

    @staticmethod
    def calc(phat, shape_l, shape_r):
        left = AMLSE.calc(phat, 'left', shape_l)
        right = AMLSE.calc(phat, 'right', shape_r)
        return tf.reduce_mean([left,right])

    def update_state(self, y, phat, sample_weight=None):
        self.two_tailed.assign(TwoTailedAMLSE.calc(phat, self.shape_l, self.shape_r))
        
    def result(self):
        return self.two_tailed

    def reset_state(self):
        self.two_tailed.assign(0)

class NLL(tf.keras.metrics.Metric):
    def __init__(self, name='nll', **kwargs):
        super(NLL, self).__init__(name=name, **kwargs)
        self.nll = self.add_weight(name='nll', initializer='zeros')

    @staticmethod
    def calc_nll(y_, axis=None):
        """
        """
        return tf.reduce_mean(y_, axis=axis)

    @staticmethod
    def calc(y, phat, axis=None):
        return NLL.calc_nll(-phat.log_prob(y), axis=axis)
        
    def update_state(self, y, phat, sample_weight=None):
        self.nll.assign(NLL.calc(y, phat))
                          
    def result(self):
        return self.nll

    def reset_state(self):
        self.nll.assign(0)

class PhatLossMetric(tf.keras.metrics.Metric):
    def __init__(self, *args, name='phat_loss_metric', **kwargs):
        super(PhatLossMetric, self).__init__(name=name, **kwargs)
        self.phat_metric = self.add_weight(name='phat_metric', initializer='zeros')
        
        self.nll = NLL()
        self.tails = TwoTailedAMLSE(*args, **kwargs)

    def update_state(self, y, phat, sample_weight=None):
        nll = self.nll.calc(y, phat)
        tails = self.tails.calc(phat, self.tails.shape_l, self.tails.shape_r)
        self.phat_metric.assign(nll / (tails + 1))
                          
    def result(self):
        return self.phat_metric

    def reset_state(self):
        self.phat_metric.assign(0)
        
class BodyLoss(tf.keras.losses.Loss):
    def __init__(self):
        super(BodyLoss, self).__init__(name="BodyLoss")
        
    def call(self, y, y_):
        return NLL.calc_nll(y_)

class TailLoss(tf.keras.losses.Loss):
    def __init__(self, shape_l, shape_r):
        super(TailLoss, self).__init__(name="TailLoss")
        self.shape_l = shape_l
        self.shape_r = shape_r
        
        self.shapes = tf.constant([[shape_l, shape_r]], dtype=tf.float64)
        
    def call(self, y, shapes_):
        shapes = tf.tile(self.shapes, tf.shape(y))
        loss = AMLSE.calc_amlse(shapes_, shapes)
        return loss
    
class PhatLoss(tf.keras.losses.Loss):
    """
    Custom loss function for PHAT model assessment
    
    Convex combination of:
    1) the negative likelihood
    2) RMSE of extreme quantitles
        > ensures fit in the tails
    """
    def __init__(self, *args, **kwargs):
        super(PhatLoss, self).__init__(name="PhatLoss")
        self.BodyLoss = BodyLoss()
        self.TailLoss = TailLoss(*args, **kwargs)

    def call(self, y, out, sample_weight=None):
        y_ = out[:,0]
        shapes_ = out[:,1:]
        body_loss = self.BodyLoss(y, y_)
        tail_loss = self.TailLoss(y, shapes_)
        return body_loss / (tail_loss + 1)

class PhatNetBeta(tf.keras.Model, GraphicMixin):
    def __init__(self, neurons=4):
        tf.keras.backend.set_floatx('float64')
        super(PhatNetBeta, self).__init__(name='PhatNet')
        
        self.h1 = tf.keras.layers.Dense(neurons, input_shape=(1,), name='h1')

        self.mu = tf.keras.layers.Dense(1, name='mu')
        self.sigma = tf.keras.layers.Dense(1, activation='softplus', name='sigma')
        self.shape_l = tf.keras.layers.Dense(1, activation='softplus', name='shape_left')
        self.shape_r = tf.keras.layers.Dense(1, activation='softplus', name='shape_right')        

        self.vec = tf.keras.layers.Concatenate(name='vec')
                
    def call(self, xy):
        x = self.h1(xy)
        
        mu_v = self.mu(x)
        sigma_v = self.sigma(x)
        shape_l_v = self.shape_l(x)
        shape_r_v = self.shape_r(x)

        vec = self.vec([mu_v, sigma_v, shape_l_v, shape_r_v])
        return vec

    @splitter
    def train_step(self, x, y, sample_weight):            
        with tf.GradientTape() as tape:
            vec = self(x, training=True)
            mean_, std_, shape_l_, shape_r_ = tf.unstack(vec, num=4, axis=-1)
        
            phat_ = Phat4TF(mean_, std_, shape_l_, shape_r_)
            y_ = NLL.calc(y, phat_, axis=1)

            loss = self.compiled_loss(
                y,
                y_,
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )
        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        self.compiled_metrics.update_state(y, phat_)

        return {m.name: m.result() for m in self.metrics}
    
    @splitter
    def test_step(self, x, y, sample_weight):
        with tf.GradientTape() as tape:
            vec = self(x, training=True)
            mean_, std_, shape_l_, shape_r_ = tf.unstack(vec, num=4, axis=-1)
        
            phat_ = Phat4TF(mean_, std_, shape_l_, shape_r_)
            y_ = NLL.calc(y, phat_, axis=1)

            loss = self.compiled_loss(
                y,
                y_,
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )
        self.compiled_metrics.update_state(y, phat_)

        return {m.name: m.result() for m in self.metrics}

    def early_stop(self,
            monitor='val_loss', 
            min_delta=0.00001,
            patience=5, 
            verbose=1, 
            mode='min', 
            *args, **kwargs
        ):
        self._early_stop = tf.keras.callbacks.EarlyStopping(
            monitor=monitor, 
            min_delta=min_delta, 
            patience=patience,
            verbose=verbose,
            mode=mode,
            *args, **kwargs
        )
    
    def fit(self, 
            *args, 
            early_stop:bool=True,
            **kwargs
        ):
        if 'callbacks' not in kwargs:
            kwargs['callbacks'] = []
        
        if not hasattr(self, '_early_stop') and early_stop:
            self.early_stop()
        
        if hasattr(self, '_early_stop'):
            kwargs['callbacks'].append(self._early_stop)

        return super().fit(*args, **kwargs)

class WeightsCallBack(tf.keras.callbacks.Callback):
    def __init__(self, weights={}):
        super(WeightsCallBack, self).__init__()  
        self.weights = weights
        
    def on_batch_end(self, batch, logs={}):
        curr_epoch = list(self.weights.keys())[-1]
        self.weights[curr_epoch].update({batch: self.model.get_weights()})

    def on_epoch_begin(self, epoch, logs={}):
        self.weights[epoch] = {}

class PhatNet(tf.keras.Model, GraphicMixin):
    PARAM_NAMES = ['mean', 'sig', 'shape_l', 'shape_r']
    def __init__(self, neurons=1):
        tf.keras.backend.set_floatx('float64')

        super(PhatNet, self).__init__(name='PhatNet')
            
        self.h1 = tf.keras.layers.Dense(neurons, input_shape=(1,), name='h1')

        self.mu = tf.keras.layers.Dense(1, name='mu')
        self.sigma = tf.keras.layers.Dense(1, activation='softplus', name='sigma')
        self.shape_l = tf.keras.layers.Dense(1, activation='softplus', name='shape_left')
        self.shape_r = tf.keras.layers.Dense(1, activation='softplus', name='shape_right')        

        self.body = tf.keras.layers.Concatenate(name='body')
        self.tails = tf.keras.layers.Concatenate(name='tails')
                
    def call(self, x):
        x = self.h1(x)

        mu_v = self.mu(x)
        sigma_v = self.sigma(x)
        shape_l_v = self.shape_l(x)
        shape_r_v = self.shape_r(x)

        body = self.body([mu_v, sigma_v])
        tails = self.tails([shape_l_v, shape_r_v])
        return body, tails
    
    @splitter
    def train_step(self, x, y, sample_weight):            
        with tf.GradientTape() as tape:
            body_, shapes_ = self(x, training=True)
            mean_, std_ = tf.unstack(body_, num=2, axis=-1)
            phat_ = Phat4TF(mean_, std_, shapes_[:,0], shapes_[:,1])
            nll = NLL.calc(y, phat_, axis=1)
            y_ = tf.concat((tf.reshape(nll, (-1,1)), shapes_), axis=1)

            loss = self.compiled_loss(
                y,
                y_,
                sample_weight=sample_weight,
                regularization_losses=self.losses,
            )
        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)
        
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        self.compiled_metrics.update_state(y, phat_)

        return {m.name: m.result() for m in self.metrics}
    
    def predicted_params(self):
        predicted = self.predict([0])
        vals = np.concatenate(list(predicted)).flatten()
        dict = {name: val for name, val in zip(self.PARAM_NAMES, vals)}
        df = pd.DataFrame([dict], index=['']).T

        return df

    @splitter
    def test_step(self, x, y, sample_weight):
        body_, shapes_ = self(x, training=True)
        mean_, std_ = tf.unstack(body_, num=2, axis=-1)

        phat_ = Phat4TF(mean_, std_, shapes_[:,0], shapes_[:,1])
        nll = NLL.calc(y, phat_, axis=1)

        y_ = tf.concat((tf.reshape(nll, (-1,1)), shapes_), axis=1)

        loss = self.compiled_loss(
            y,
            y_,
            sample_weight=sample_weight,
            regularization_losses=self.losses,
        )

        self.compiled_metrics.update_state(y, phat_)
        return {m.name: m.result() for m in self.metrics}
    
    def early_stop(self,
            monitor='val_loss', 
            min_delta=0.0001,
            patience=5, 
            verbose=1, 
            mode='min', 
            *args, **kwargs
        ):
        self._early_stop = tf.keras.callbacks.EarlyStopping(
            monitor=monitor, 
            min_delta=min_delta, 
            patience=patience,
            verbose=verbose,
            mode=mode,
            *args, **kwargs
        )
    
    def fit(self, 
            *args, 
            weights:dict=None, 
            logdir:str='',
            early_stop:bool=True,
            nan_stop:bool=False,
            **kwargs
        ):
        if 'callbacks' not in kwargs:
            kwargs['callbacks'] = []
        
        if not hasattr(self, '_early_stop') and early_stop:
            self.early_stop()
        
        if hasattr(self, '_early_stop'):
            kwargs['callbacks'].append(self._early_stop)

        if nan_stop:
            kwargs['callbacks'].append(tf.keras.callbacks.TerminateOnNaN())

        if isinstance(weights, dict):
            kwargs['callbacks'].append(WeightsCallBack(weights))

        if logdir:
            kwargs['callbacks'].append(tf.keras.callbacks.TensorBoard(logdir, histogram_freq=1))

        return super().fit(*args, **kwargs)



"""
Gaussian-based neural network
"""
import tensorflow as tf
from tensorflow_probability import distributions as tfd

from phat.learn.utils import GraphicMixin, nnelu

def mon_mean(y, y_):
    mu, sigma = tf.unstack(y_, num=2, axis=-1)
    return tf.keras.backend.mean(mu)

def mon_std(y, y_):
    mu, sigma = tf.unstack(y_, num=2, axis=-1)
    return tf.keras.backend.mean(sigma)

def gnll_loss(y, y_):
    """ Computes the mean negative log-likelihood loss of y given the mixture parameters.
    """
    mu, sigma = tf.unstack(y_, num=2, axis=-1)
    dist = tfd.Normal(mu, sigma)
    return tf.reduce_mean(-dist.log_prob(y))

class DN(tf.keras.Model, GraphicMixin):
    def __init__(self, neurons=200):
        tf.keras.backend.set_floatx('float64')
        tf.keras.utils.get_custom_objects().update({
            'nnelu': tf.keras.layers.Activation(nnelu),
            'mean': mon_mean,
            'std': mon_std
        })
        super(DN, self).__init__(name="DN")
        self.neurons = neurons
        
        self.h1 = tf.keras.layers.Dense(neurons, input_shape=(1,), activation='tanh', name='h1')
        
        self.mu = tf.keras.layers.Dense(1, name="mu")
        self.sigma = tf.keras.layers.Dense(1, activation="nnelu", name="sigma")
        self.pvec = tf.keras.layers.Concatenate(name="pvec")

    def call(self, inputs):
        x = self.h1(inputs)
        
        mu_v = self.mu(x)
        sigma_v = self.sigma(x)
        
        return self.pvec([mu_v, sigma_v])



"""
Houses all scipy-based distributions.

Separate code required for Tensorflow. Found in .learn.dists

WARNING: Only CarBen and Phat distributions should be used. The remaining
distributions are for illustration purposes in the docs.
"""
from typing import Union, Iterable, Sequence, Callable, List
import numpy as np
import scipy.stats as scist
import scipy.special as spec
from scipy._lib._util import check_random_state
import statsmodels.api as sm
from statsmodels.base.model import GenericLikelihoodModel

from phat.utils import argsetter, arrayarize, sech_squared, stacker, dotweight, XTYPE

class BLineSig:
    """
    Borderline Sigmoid function defined by Taleb in SCOFT https://arxiv.org/abs/2001.10488

    FOR ILLUSTRATION PURPOSES ONLY
    """
    def __init__(self, k:float=1):
        self.k = k

    @argsetter('xk')
    def sf(self, x, k:float=None):
        return .5*(1 - np.tanh(k*x))

    @argsetter('xk')
    def pdf(self, x, k:float=None):
        return .5*k*sech_squared(x,k)
    
    @argsetter('k')
    def var(self, k):
        return np.pi**2 / (12*k**2)
    
    @argsetter('k')
    def m4(self, k):
        return 7*np.pi**4 / (240*k**4)
    
    @argsetter('k')
    def kurt(self, k):
        return self.m4(k) / self.var(k)**2

class LogPareto:
    """
    LogPareto as discussed in SCOFT https://arxiv.org/abs/2001.10488.

    FOR ILLUSTRATION PURPOSES ONLY
    """
    def __init__(self, alpha=2, L=1):
        self.alpha = alpha
        self.L = L
        
    def sf(self, x):
        return (self.L**self.alpha)*(np.log(x)**(-self.alpha))

class DBLFat(scist.rv_continuous):
    """
    Baseclass for group of twin-tailed fat-tailed distributions

    FOR ILLUSTRATION PURPOSES ONLY    
    """
    @property
    def dist(self):
        raise NotImplementedError

    @property
    def qjunc(self):
        return .5

    @argsetter(['x', 'shape'])
    def _pdf(self, x, shape):
        pdf = np.where(
            x>=0, 
            self.dist.pdf(x, shape), 
            self.dist.pdf(-x, shape)
        )
        return pdf / 2
    
    @argsetter(['x', 'shape'])
    def _cdf(self, x, shape):
        cdf = np.where(
            x>=0,
            .5 + (self.dist.cdf(x, shape)/2), 
            
            self.dist.sf(-x, shape)/2
        )
        return cdf
    
    @argsetter(['q', 'shape'])
    def _ppf(self, q, shape):
        if isinstance(q, (float, int)):
            q = np.array([q])
        
        ppf = np.where(
            q>=self.qjunc,
            self.dist.ppf(1/self.qjunc*(q - self.qjunc), shape),
            -self.dist.isf(q*1/self.qjunc, shape)
        )

        return ppf
    
    def _rvs(self, *args, **kwargs):
        rvs = super()._rvs(*args, **kwargs)
        return rvs
    
class DBLGP(DBLFat):
    dist = scist.genpareto
    
class DBLLomax(DBLFat):
    dist = scist.lomax
    
class DBLPareto(DBLFat):
    dist = scist.pareto

############# Main Phat-related Distributions ################
class CarBenBase:
    """
    Base Class of CarBen Hybrid distribution. Generalizes code used in both Left and Right tailed versions.

    CarBen is a one-tailed assymetric hybrid of Gaussian and generalized Pareto distributions. Derivation of 
    the distribution found in Carreau and Bengio (2008):
    https://www.researchgate.net/publication/226293435_A_hybrid_Pareto_model_for_asymmetric_fat-tailed_data_The_univariate_case#pf6

    Takes 3 free parameters:
        > shape (xi) of generalized Pareto (inverse of the tail index)
        > mean (mu), standard deviation (sig) of the Gaussian distribution
    
    Remaining 2 Pareto parameters are found via equalities
        > b, which is the scale, as function of sig and xi (utilizing a Lambert W)
        > a, which is the location, can then be found either via a) scale OR b) via the Lambert W. the latter is used
    
    Optional parameters:
        > instead of `sig`, user can provide `a`, which is the location of the generalized Pareto

    Parameters
    -----------
    xi:     inverse of tail index of generalized Pareto tail
    mu:     mean of the Gaussian body
    sig:    standard deviation of the Gaussian body

    Attributes
    -----------
    body:   scipy gaussian distribution with arguments mean/loc=mu and scale=sig
    tail:   scipy generalized pareto with arguments shape=xi, loc=a, scale=b
    gamma:  the proportionate factor whose inverse ensures pdf of CarBen integrates to 1

    Methods
    --------
        > handle helper and common distribution functions that are agnostic to tail-direction
    """

    def __init__(self, 
        xi:float,
        mu:float,
        sig:float=None, 
        a:float=None
        ) -> None:
        """
        Calcs sig, a, b paramters based on arguments provided

        Assigns body distribution with given arguments. The tail distribution is assigned in child classes.
        """
        self.xi = xi
        self.mu = mu

        provided = (sig is None, a is None)
        if all(provided) or not any(provided):
            raise ValueError('Please provide only one of `sig` or `a`')
        elif sig is not None:
            self.sig = sig
            self._given = 'sig'
            self.b = self._calc_b_w_sig()
            self.a = self._calc_a()
        elif a is not None:
            self.a = a
            self._given = 'a'
            self.b = self._calc_b_w_a()
            self.sig = self._calc_sig()

        self.body = scist.norm(self.mu, self.sig)
        
    @property
    def tail(self) -> str:
        """
        Tail distribution of type scipy.genpareto.

        `_tail` is assigned in CarbenLeft or CarbenRight child class
        """
        if hasattr(self, '_tail'):
            return self._tail
        else:
            raise NotImplementedError

    @property
    def gamma(self) -> float:
        """
        defined in Carreau and Bengio (2008)

        proportionate factor whose inverse scales the CarBen so the density integrates to 1
        """
        val = np.sqrt(self.W_z() / 2)
        return 1 + .5*(1 + spec.erf(val))

    def z(self) -> float:
        """
        in Carreau and Bengio (2008), z defined as:

            (1 + xi)**2 / 2*pi
        """
        num = (1 + self.xi)**2
        denom = 2*np.pi
        return  num / denom        

    def W_z(self) -> float:
        """
        in Carreau and Bengio (2008), W(z) can be calculated directly as:
            
            sig**2 * (1 + xi)**2 / b**2

        b is not a free parameter, so W(z) must be calculated via
        the Lambert function, which is defined only in free parameters.
        """

        val = spec.lambertw(self.z())
        assert (val.imag == 0).all(), val.imag

        return val.real

    def _calc_b_w_sig(self) -> float:
        """
        see Equation 7 in Carreau and Bengio (2008)
        """
        num = self.sig*(1 + self.xi)
        denom = np.sqrt(self.W_z())
        
        return num / denom

    def _calc_b_w_a(self) -> float:
        """
        found by substituion of Equation 7 in Carreau and Bengio (2008)

        sig = (a - mu)/W(z)
        b = ((1+xi)(a - mu)/(W(z)sqrt(W(z)))
        b = a - mu + xi*a - xi*mu
        """
        num = self.a - self.mu*self.xi
        denom = np.sqrt(self.W_z())
        return num / denom

    @argsetter('x')
    def sf(self, x:XTYPE=None) -> np.ndarray:
        """
        Survival function, P(X < x)

        Parameters
        ----------
        x:      float, m x n iterable of floats of random variable samples

        Return
        -------
        float or iterable of floats returning survival probability of each sample
        """
        return 1 - self.cdf(x)
    
    def rvs(self, 
        size:Union[int, Iterable[Sequence[int]]]=None, 
        seed:int=None) -> XTYPE:
        """
        Generates random draws

        1. Draws randomly from [0,1] uniform distribution
        2. Calls quantile function for each draw

        Mimics scipy process:
        https://github.com/scipy/scipy/blob/28e4811d48c99b7c68e41992eda5cff859c1fa2b/scipy/stats/_distn_infrastructure.py#L1032

        Params
        -------
        size:   int or iterable of ints providing dimensions of return array
        seed:   int; initialize with specific random state for replication
        """
        random_state = check_random_state(seed)

        U = random_state.uniform(size=size)
        if size is None:
            U = np.array(U, ndmin=1)
            Y = self.ppf(U)[0]
        else:
            Y = self.ppf(U)
        return Y

class CarBenRight(CarBenBase):
    """
    Right-tailed CarBen distribution.

    Assigns `_tail`, handler methods, and common distribution functions that 
    are specific to the tail direction.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tail = scist.genpareto(self.xi, loc=self.a, scale=self.b)
    
    def _calc_a(self) -> float:
        """
        see Equation 8 in Carreau and Bengio (2008)        
        """
        return self.mu + self.sig*np.sqrt(self.W_z())

    def _calc_sig(self) -> float:
        """
        rearrangement of Equation 8 in Carreau and Bengio (2008)        
        """
        return (self.a - self.mu) / np.sqrt(self.W_z())

    @argsetter('x')
    def pdf(self, x:XTYPE=None) -> np.ndarray:
        """
        Probability density function

        See Carreau and Bengio (2008) for derivation

        Parameters
        ----------
        x:      float, m x n iterable of floats of random variable samples

        Return
        -------
        float or iterable of floats returning likelihood of occurence of each sample
        """        
        pdf = np.where(
            x > self.a,
            self.tail.pdf(x),
            self.body.pdf(x)
        )
        return pdf / self.gamma
    
    @argsetter('x')
    def cdf(self, x:XTYPE=None) -> np.ndarray:
        """
        Cumulative Distribtion Function, P(X<x)

        See Docs for derivation

        Parameters
        ----------
        x:      float, m x n iterable of floats of random variable samples

        Return
        -------
        float or iterable of floats returning cumulative probability of each sample
        """
        cdf = np.where(
            x > self.a,
            self.body.cdf(self.a) + self.tail.cdf(x),
            self.body.cdf(x)
        )
        return cdf / self.gamma 

    @property
    def qjunc(self) -> float:
        """
        Quantile of junction point, a, b/w body and tail distribution
        
        See Docs for derivation
        """
        return 1 - 1 / self.gamma

    def ppf(self, q:XTYPE) -> np.ndarray:
        """
        Quantile function, analogous to `ppf` method in `scipy`.

        See Docs for derivation

        Parameters
        ----------
        q:      float, iterable of floats between [0, 1]

        Return
        -------
        float or iterable of floats returning x values corresponding to the given quantiles
        """
        ppf = np.where(
            q > self.qjunc,
            self.tail.ppf(self.gamma * (q - self.qjunc)),
            self.body.ppf(self.gamma*q)
        )
        return ppf    

    def mean(self) -> float:
        """
        First moment

        See Docs for derivation
        """
        bmu = scist.truncnorm(
            -np.inf, 
            self.a, 
            *self.body.args
        ).mean()
        tmu = self.tail.mean()
        rmu = (bmu + tmu) / self.gamma

        return rmu

    def var(self) -> float:
        """
        Second moment

        See Docs for derivation
        """
        bvar = scist.truncnorm(
            -np.inf, 
            self.a, 
            *self.body.args
        ).var()
        tvar = self.tail.var()
        rvar = (bvar + tvar) / self.gamma
        return rvar

class CarBenLeft(CarBenBase):
    """
    Right-tailed CarBen distribution.

    Assigns `_tail`, handler methods, and common distribution functions that 
    are specific to the tail direction.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        
        # NOTE: `a` is inverted to reflect the distribution appropriately
        self._tail = scist.genpareto(self.xi, loc=-self.a, scale=self.b)

    def _calc_a(self) -> float:
        """
        see Equation 8 in Carreau and Bengio (2008)   

        For the left-tail, the valence of second term is changed to account for 
        reflection of the distribution.
        """
        return self.mu - self.sig*np.sqrt(self.W_z())

    def _calc_sig(self) -> float:
        """
        rearrangement of Equation 8 in Carreau and Bengio (2008)        

        Again, the valence is changed
        """
        return (self.mu - self.a) / np.sqrt(self.W_z())

    @argsetter('x')
    def pdf(self, x:XTYPE=None) -> np.ndarray:
        """
        Probability density function

        See Docs for derivation
        
        Parameters
        ----------
        x:      float, m x n iterable of floats of random variable samples

        Return
        -------
        float or iterable of floats returning the likelihood of occurence of each sample        
        """
        pdf = np.where(
            x < self.a,
            self.tail.pdf(-x),
            self.body.pdf(x)            
        )
        return pdf / self.gamma
    
    @argsetter('x')
    def cdf(self, x:XTYPE=None) -> np.ndarray:
        """
        Cumulative Distribtion Function, P(X<x)

        See Docs for derivation

        Parameters
        ----------
        x:      float, m x n iterable of floats of random variable samples

        Return
        -------
        float or iterable of floats returning cumulative probability of each sample
        """
        cdf = np.where(
            x < self.a,
            (1 - self.tail.cdf(-x)),
            1 + self.body.cdf(x) - self.body.cdf(self.a),
        )
        return cdf / self.gamma
        
    @property
    def qjunc(self):
        """
        Quantile of the juncture point, a, between the tail and body
        
        See Docs for derivation
        """
        return 1 / self.gamma
    
    def ppf(self, q:XTYPE) -> np.ndarray:
        """
        Quantile function, analogous to `ppf` method in `scipy`.

        See Docs for derivation

        Parameters
        ----------
        q:      float, iterable of floats between [0, 1]

        Return
        -------
        float or iterable of floats returning x values corresponding to the given quantiles
        """
        ppf = np.where(
            q < self.qjunc,
            -self.tail.ppf(1 - q*self.gamma),
            self.body.ppf(self.gamma*q + self.gamma*self.qjunc - self.gamma)
        )
        return ppf

    def mean(self) -> float:
        """
        First moment

        See Docs for derivation
        """
        bmu = scist.truncnorm(
            self.a,
            np.inf, 
            *self.body.args
        ).mean()
        tmu = -self.tail.mean()
        lmu = (bmu + tmu) / self.gamma

        return lmu

    def var(self) -> float:
        """
        Second moment

        See Docs for derivation
        """
        bvar = scist.truncnorm(
            self.a,
            np.inf, 
            *self.body.args
        ).var()
        tvar = self.tail.var()
        lvar = (bvar + tvar) / self.gamma
        return lvar

class CarBenHybrid: 
    """
    Light wrapper that simply reads the args/kwargs and
    selects the tail to be used, either CarBenLeft or CarBenRight

    If `xi>=0`:
        if `rtail` argument == True, use CarBenRight
        if `rtail` argument == False, use CarBenLeft
    elif `xi<0`:
        use CarBenLeft

    i.e. You can simply pass a negative tail index to return a left-skewed CarBen
    """  
    def __new__(cls, *args, **kwargs):
        """
        Takes arguments as per CarBenLeft/CarBenRight/CarBenBase

        Returns 
        --------
        CarBenLeft or CarBenRight object
        """
        args = list(args)
        xi_is_arg =  len(args) >= 0
        xi = args[0] if xi_is_arg else kwargs['xi']
        
        rtail_is_arg = len(args) >= 5
        if rtail_is_arg:
            rtail = args.pop(4)
        else:
            rtail = kwargs.pop('rtail') if 'rtail' in kwargs else True

        if isinstance(xi, float):
            if xi < 0 and not rtail:
                txt = 'If you provide a negative xi parameter,'
                txt += ' do not provide `rtail`.'
                txt += ' A left-tailed CarBen is computed automatically.'
                raise ValueError(txt)             
            elif xi < 0:
                rtail = False
                if xi_is_arg:
                    args[0] = -xi
                else:
                    kwargs['xi'] = xi

        if rtail:
            obj = CarBenRight(*args, **kwargs)
        else:
            obj = CarBenLeft(*args, **kwargs)
            
        return obj

class PhatFit(GenericLikelihoodModel):
    """
    Light wrapper for statsmodels GenericLogLikelihood. Used to mirror scipy or 
    statsmodels fit() method.

    Parameters
    -----------
    endog:      iterable of dependet variable
    exog:       iterable of independet variable
    xi_l:       inverse of tail index of left-tailed generalized Pareto tail. Default: None
    xi_r:       inverse of tail index of right-tailed generalized Pareto tail. Default: None
    tail_est:   func; process used to estimate both left and right tail indices. Default: None

    Attributes
    -----------
    left:               left-tailed CarBenHybrid object
    right:              right-tailed CarBenHybrid object
    args:               list of values of 8 parameters of Phat distribution
    params:             name-value pairs of 8 parameters of Phat distribution
    learnable_params:   name-value pairs of 4 given parameters of Phat distribution

    Methods
    --------
        > handle helper and common distribution functions that are agnostic to tail-direction
    """
    def __init__(self, 
        endog:Iterable, 
        exog:Iterable,
        xi_left:float=None, 
        xi_right:float=None, 
        tail_est:Callable=None,
        **kwds):
        
        if xi_right is None and xi_left is None and tail_est is not None:
            self.xi_left, self.xi_right = tail_est(endog)
        else:
            self.xi_left, self.xi_right = xi_left, xi_right
            
        super().__init__(endog, exog, **kwds)

    def nloglikeobs(self, params):
        """
        Adjust Phat object parameters based on arguments provided at instantiation
        """
        if len(params) == 4:
            phatdist = Phat(*params)
        else:
            phatdist = Phat(params[0], params[1], self.xi_left, self.xi_right)
        return phatdist.nll(self.endog)

    def fit(self, start_params=None, maxiter=10000, maxfun=5000, **kwds):
        """
        Determines start parameters of fit method based on arguments provided at instantiation

        start_params for Gaussian are set via scipy `norm` distribution.
        """
        if start_params == None:
            if self.xi_left is None and self.xi_right is None:
                self.exog_names.append('xi_l')
                self.exog_names.append('xi_r')
                start_params = np.zeros(4)
                start_params[:2] = scist.norm.fit(self.endog)
                start_params[-2:] = [.2, .2]
            else:
                start_params = np.zeros(2)
                start_params = scist.norm.fit(self.endog)
            
        return super().fit(
            start_params=start_params,
            maxiter=maxiter, maxfun=maxfun,
            **kwds
        )

class Phat:
    """
    Twin-tailed distribution that combines a right-tailed CarBen with a left-tailed CarBen

    Parameters:
        mu:    
            mean of the Gaussian body
        sig:
            standard deviation of the Gaussian body
        xi_l:   
            inverse of tail index of left-tailed generalized Pareto tail
        xi_r:   
            inverse of tail index of right-tailed generalized Pareto tail
        p:
            relative weights of the two CarBen hybrids. Default: [.5, .5]

    Attributes:
        left:               
            left-tailed CarBenHybrid object
        right:              
            right-tailed CarBenHybrid object

    **Details**

    Per Carreau and Bengio (2008).
    
    The properites of the body in both distributions are equivalent; thus, the mixture requires just
    four parameters: two parameters for the Gaussian body and the tail index for each of the two tails. 
    In all, the distribution has 8 parameters:

    + Left tail: xi, a, b
    + Right tail: xi, a, b
    + Center: mu, sig

    As with any mixture model, this mix is the weighted average result of values from the two
    components. Default weights are 0.5 / 0.5.

    Weights could otherwise be estimated via machine learning, threshold analysis (for the tails),
    and gaussian statistics (for the body).
    """
    PARAM_NAMES = ['mu', 'sig', 'xi_l',  'xi_r', 'a_l', 'a_r', 'b_l', 'b_r']

    def __init__(
        self, mu:float, sig:float, xi_l:float, xi_r:float, p:Iterable=None
        ) -> None:
        """
        Left and right tails are instantiated immediately
        """
        self.mu = mu
        self.sig = sig
        self.xi_l = xi_l if xi_l >= 0 else -xi_l
        self.xi_r = xi_r

        self.left = CarBenHybrid(xi_l, mu, sig, rtail=False)
        self.right = CarBenHybrid(xi_r, mu, sig)

        self.p = p if p is not None else np.array([.5,.5])
    
    @property
    def args(self) -> tuple:
        """
        Tuple of the distribution parameters
        """
        return self.mu, self.sig, self.xi_l, self.xi_r, self.left.a, \
            self.right.a, self.left.b, self.right.b

    @property
    def params(self) -> dict:
        """
        Dictionary of the distribution parameters
        """      
        return {name: arg for name, arg in zip(self.PARAM_NAMES, self.args)}

    @property
    def learnable_params(self) -> List[float]:
        """
        List of distribution parameters that can be tuned by machine learning approach
        """
        return [self.params[name] for name in self.PARAM_NAMES[:4]]

    @argsetter('x')
    @dotweight
    def pdf(self, x:XTYPE=None) -> np.ndarray:
        return stacker(x)((self.left.pdf(x), self.right.pdf(x)))
    
    @argsetter('x')
    @dotweight    
    def cdf(self, x:XTYPE=None) -> np.ndarray:
        return stacker(x)((self.left.cdf(x), self.right.cdf(x)))
    
    @argsetter('x')
    def sf(self, x:XTYPE=None) -> np.ndarray:
        return 1 - self.cdf(x)
    
    @dotweight
    def ppf(self, q:XTYPE=None) -> np.ndarray:
        """
        Parameters:
            q: float, iterable of floats 
                Between [0, 1]
        """        
        return stacker(q)((self.left.ppf(q), self.right.ppf(q)))

    def loglike(self, x=None):
        "Loglikelihood"
        return np.log(self.pdf(x))

    @argsetter('x')
    def nll(self, x:XTYPE=None) -> np.ndarray:
        """Negative loglikelihood used in machine learning cost functions"""
        return -np.log(self.pdf(x))

    def rvs(self, 
        size:Union[int, Iterable[Sequence[int]]]=None, 
        seed:int=None) -> XTYPE:
        """
        Generates random draws

        **Process**

        1. Draws randomly from [0,1] uniform distribution
        2. Calls quantile function for each draw

        Mimics `scipy process <https://github.com/scipy/scipy/blob/28e4811d48c99b7c68e41992eda5cff859c1fa2b/scipy/stats/_distn_infrastructure.py#L1032>`_.
        """
        random_state = check_random_state(seed)
        U = random_state.uniform(size=size)
        if size is None:
            U = np.array(U, ndmin=1)
            Y = self.ppf(U)[0]
        else:
            Y = self.ppf(U)
        
        return Y

    @dotweight
    def mean(self) -> float:
        return np.array([self.left.mean(), self.right.mean()])

    @dotweight
    def var(self) -> float:
        return np.array([self.left.var(), self.right.var()])

    def std(self) -> float:
        return np.sqrt(self.var())

    def std_rvs(self, *args, **kwargs) -> XTYPE:
        """
        Generates random variables standardized by dividing by the standard deviation

        Resulting samples should have a standard deviation of 1
        """
        return self.rvs(*args, **kwargs) / self.std()

    @staticmethod
    def fit(values, xi_l:float=None, xi_r:float=None) -> PhatFit:
        """
        Generates standard log-likelihood fit to Phat distribution

        Returns:
            PhatFit
        """
        exog = sm.add_constant(np.zeros_like(values), prepend=True)
        phatfit = PhatFit(values, exog, xi_l, xi_r)

        return phatfit.fit()  

class PhatStack(Phat):
    """
    Weighted model of multiple Phat distributions. 
    
    This is NOT a mixture model. Used only to generate a two dimensional set
    of random variables that are consistent in a single component
    distribution across the second dimension.
    
    Parameters
    -----------
    phats:  iterable or args Phat distributions
    p:      iterable of floats; represents weighting of each Phat distribution
            in the final model. Must sum to 1.

    Attributes
    -----------
    n_comps:      int; number of component phat distributions
    
    Methods
    --------
    rvs
    
    """    
    def __init__(
        self, 
        *phats:Phat,
        p:Iterable=None
        ) -> None:
        """
        """
        if len(phats) == 1:
            phats = phats[0]
        
        if len(phats) != len(p):
            txt = 'Length of probabilities does not match number' 
            txt += f' of distributions: {len(phats)} v {len(p)}'
            raise AssertionError(txt)
        
        self.n_comps = len(phats)
        self.phats = phats
        self.p = arrayarize(p)

    def rvs(self, 
        size:Iterable[int]=None, 
        seed:int=None,
        return_splits:bool=False,
        ) -> XTYPE:
        """
        Generates array of random draws.
        
        ***IMPORTANT***
        Accepts only 2D m x n array. Draws are taken proportionately from
        the component Phat distributions ONLY along the first axis. Along the 
        second axis, draw are taken ONLY from the specific Phat distribution 
        isolated in the first axis.
        
        This ensures that only one distribution regime impacts the time series
        over time.
        
        Where the required iterations, m, does not divide evenly into the
        proportions, the remainders are added to certain distributions

        Params
        -------
        size:           iterable; m x n. Only two dimensions allowed.
        seed:           int; initialize with specific random state for replication
        return_splits:  bool; optional return of the number of m iterations
                        returned for each component distribution
        """
        if len(size) != 2:
            raise AssertionError('Size must be 2D m x n array')

        iters = np.floor(size[0] * self.p).astype(int)
        rmdr = size[0] - iters.sum()
        iters[:rmdr] += np.ones(rmdr, dtype=int)
        
        Y = np.zeros(size)
        st = 0
        for i, end in enumerate(iters):
            Y[st: st+end] = self.phats[i].rvs((end, size[1]), seed=seed)
            st += end
        
        if return_splits:
            return Y, iters
        else:
            return Y
            


from functools import wraps
from typing import Iterable

import numpy as np
import scipy.stats as scist
import matplotlib.pyplot as plt

from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FloatVector

from phat.utils import argsetter

base = importr('base')
utils = importr('utils')
utils.chooseCRANmirror(ind=1)
utils.install_packages('POT')
POT = importr('POT')

def fit_line_within(stacked, ival):
    ivalmask = np.logical_and(stacked[:,0]>=ival[0], stacked[:,0]<=ival[1])
    return (*scist.linregress(stacked[ivalmask])), ivalmask.sum()

def threshset(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        data = kwargs['data']
        spacer = 45 if 'spacer' not in kwargs else kwargs['spacer']
    
        if not hasattr(self, 'tholds') and 'tholds' not in kwargs:
            step = np.quantile(data, .995)/spacer
            tholds = np.arange(-.1, max(data), step=step)
            self.tholds = tholds
        elif 'tholds' not in kwargs:
            tholds = self.tholds
        else:
            self.tholds = kwargs['tholds']

        kwargs['tholds'] = tholds
    
        return func(self, *args, **kwargs)
    
    return wrapper


class Threshold:
    def __init__(self, data):
        self.data = data
        
    @argsetter(['data'])
    @threshset
    def MRL(self, data:Iterable=None,
            tholds:Iterable=None, alpha:float=.05,
            fig=None, ax=None, show_plot:bool=True,
            splits:Iterable=None, *args, **kwargs
        ):
        is_excess = np.array([data > thold for thold in tholds])
        excesses = np.array([data - thold for thold in tholds])
        excesses = np.where(
            is_excess,
            excesses,
            np.nan
        )

        self.mean_exc = np.nanmean(excesses, axis=1)
        stds = np.nanstd(excesses, axis=1)
        z_inverse = scist.norm.ppf(1-(alpha/2))
        CI = z_inverse*stds/(len(excesses)**0.5)
        
        if show_plot:
            if fig is None or ax is None:
                fig, ax = plt.subplots(1,1,figsize=(10,6))
            
            ax.plot(tholds, self.mean_exc)
            ax.fill_between(tholds, self.mean_exc - CI, self.mean_exc + CI, alpha = 0.4)
            ax.set_xlabel('u')
            ax.set_ylabel('Mean Excesses')
            ax.set_title('Mean Residual Life Plot')

        if splits is not None:
            self._MRL_regrs(splits, ax)

    def _MRL_regrs(self, splits:Iterable, ax):
        splits = np.array(splits)
        stacked = np.vstack([self.tholds, self.mean_exc]).T
        
        sgmts = np.vstack((splits[:-1], splits[1:])).T
        for i in range(sgmts.shape[0]):
            sgmt = sgmts[i]
            b, a, r, p, stderr, n = fit_line_within(stacked, sgmt)

            count = (self.data>sgmt[1]).sum()
            y = b*sgmt + a
            label = '[{:.4f},{:.4f}] N<{}; N>{}'.format(*sgmt, n, count) + r' $R^2: $' + f'{r**2:.0%}'
            label += f' p-value: {p:.02f}'
            ax.plot(sgmt, y, label=label)

        plt.legend(loc='best')

    @argsetter(['data'])
    @threshset
    def param_stable(self, data:Iterable=None,
            tholds:Iterable=None, alpha:float=.05,
            fig=None, axs=None,
            *args, **kwargs
        ):
        shape = []
        scale = []
        mod_scale = []
        CI_shape = []
        CI_mod_scale = []
        z = scist.norm.ppf(1-(alpha/2)) 
        for thold in tholds:
            fit, _, _ = self.fit(data=data, thold=thold.item(), est='mle')

            shape.append(fit[0][1])
            CI_shape.append(fit[1][1]*z)

            scale.append(fit[0][0])
            mod_scale.append(fit[0][0] - (fit[0][1]*thold))
            Var_mod_scale = (fit[3][0] - (thold*fit[3][2]) - thold*(fit[3][1] - (fit[3][3]*thold))) 
            CI_mod_scale.append((Var_mod_scale**0.5)*z)

        #Plotting shape parameter against u vales   
        axs[0].errorbar(tholds, shape, yerr = CI_shape, fmt = 'o' )
        axs[0].set_xlabel('u')
        axs[0].set_ylabel('Shape')
        axs[0].set_title('Shape Parameter Stability')

        #Plotting modified scale parameter against u values
        axs[1].errorbar(tholds, mod_scale, yerr = CI_mod_scale, fmt = 'o')
        axs[1].set_xlabel('u')
        axs[1].set_ylabel('Modified Scale')
        axs[1].set_title('Modified Scale Parameter Stability')

    @argsetter(['data'])
    def fit(self, data:Iterable=None, thold:float=0, est:str='mle'):
        rdata = np.sort(data)   
        data_over_thresh = rdata[rdata > thold]
        data_exc= data_over_thresh - thold

        rdata = FloatVector(rdata)
        fit = POT.fitgpd(rdata, thold, est=est)
        
        return fit, data_over_thresh, data_exc

    @argsetter(['data'])
    def qqplot(self, 
        data:Iterable=None, thold:float=0, est:str='mle', alpha:float=.05,
        fig=None, ax=None
        ):
        fit, over_thresh, _ = self.fit(data=data, thold=thold, est=est)
        scale, shape = fit[0][0], fit[0][1]

        p = []
        n = len(data)
        data = np.sort(data)  
        i_initial = np.searchsorted(data, thold)
        k = i_initial - 1

        p = (np.arange(i_initial, n) - .35) / n
        p0 = (k - 0.35)/(n)  
        quantiles = thold + ((scale/shape)*(((1-((p-p0)/(1-p0)))**-shape) - 1))

        n = len(over_thresh)
        y = np.arange(1,n+1)/n

        #Kolmogorov-Smirnov Test for getting the confidence interval
        K = (-0.5*np.log(alpha/2))**0.5
        M = (len(p)**2/(2*len(p)))**0.5
        CI_qq_high = []
        CI_qq_low = []
        for prob in y:
            F1 = prob - K/M
            F2 = prob + K/M
            CI_qq_low.append(thold + ((scale/shape)*(((1-((F1)/(1)))**-shape) - 1)))
            CI_qq_high.append(thold + ((scale/shape)*(((1-((F2)/(1)))**-shape) - 1)))
        
        a, b, r_value, p_value, std_err = scist.linregress(quantiles, over_thresh)
        ax.scatter(quantiles, over_thresh)
        
        x = np.linspace(0,1,101)*100
        ax.plot(x, a*x + b, c='black', label='Regression')
        
        ax.plot(over_thresh, CI_qq_low, linestyle='--', color='red', alpha = 0.5, lw = 0.8, label='Confidence Bands')
        ax.plot(over_thresh, CI_qq_high, linestyle='--', color='red', alpha = 0.5, lw = 0.8)
        ax.set_xlabel('Theoretical GPD Quantiles')
        ax.set_ylabel('Sample Quantiles')

        ax.legend()
        ax.set_title('Q-Q Plot')

    @argsetter(['data'])
    def ppplot(self, data:Iterable=None, thold:float=0, est:str='mle', alpha:float=.05,
        fig=None, ax=None
        ):
        fit, over_thresh, _ = self.fit(data=data, thold=thold, est=est)
        scale, shape = fit[0][0], fit[0][1]
        
        n = len(over_thresh)
        y = np.arange(1,n+1)/n
        cdf_pp = scist.genpareto.cdf(over_thresh, shape, loc=thold, scale=scale)
        
        #Getting Confidence Intervals using the Dvoretzky–Kiefer–Wolfowitz method
        data = np.sort(data)
        i_initial = np.searchsorted(data, thold)
        
        F1 = []
        F2 = []
        for i in range(i_initial, len(data)):
            e = (((np.log(2/alpha))/(2*len(over_thresh)))**0.5)
            F1.append(y[i-i_initial] - e)
            F2.append(y[i-i_initial] + e)

        ax.scatter(y, cdf_pp)
        a, b, r_value, p_value, std_err = scist.linregress(y, cdf_pp)
        ax.plot(y, a*y + b, c='black', label='Regression')
        
        ax.plot(y, F1, linestyle='--', color='red', alpha = 0.5, lw = 0.8, label = 'Confidence Bands')
        ax.plot(y, F2, linestyle='--', color='red', alpha = 0.5, lw = 0.8)
        
        ax.set_xlabel('Empirical Probability')
        ax.set_ylabel('Theoritical Probability')
        ax.legend()
        ax.set_title('P-P Plot')

    @argsetter(['data'])
    def return_value(self, data:Iterable=None, thold:float=0, alpha:float=.05,
        block_size:int=252, return_period:int=252*100, est:str='mle',
        fig=None, ax=None
        ):
        data = np.sort(data) 
        fit, over_thresh, _ = self.fit(data=data, thold=thold, est=est)
        scale, shape = fit[0][0], fit[0][1]  
        
        #Computing the return value for a given return period with the confidence interval estimated by the Delta Method
        m = return_period
        Eu = len(over_thresh)/len(data)
        x_m = thold + (scale/shape)*(((m*Eu)**shape) - 1)

        #Solving the delta method    
        d = Eu*(1-Eu)/len(data)
        e = fit[3][0]
        f = fit[3][1]
        g = fit[3][2]
        h = fit[3][3]
        a = (scale*(m**shape))*(Eu**(shape-1))
        b = (shape**-1)*(((m*Eu)**shape) - 1)
        c = (-scale*(shape**-2))*((m*Eu)**shape - 1) + (scale*(shape**-1))*((m*Eu)**shape)*np.log(m*Eu)
        CI = (scist.norm.ppf(1-(alpha/2))*((((a**2)*d) + (b*((c*g) + (e*b))) + (c*((b*f) + (c*h))))**0.5))

        ny = block_size
        N_year = return_period/block_size
        i_initial = np.searchsorted(data, thold)

        p = np.arange(i_initial,len(data))/(len(data))
        N = 1/(ny*(1 - p))
        year_array = np.arange(min(N), N_year+0.1, 0.1)
        
        #Algorithm to compute the return value and the confidence intervals for plotting
        z_N = []
        CI_z_N_high_year = []
        CI_z_N_low_year = [] 
        for year in year_array:
            z_N.append(thold + (scale/shape)*(((year*ny*Eu)**shape) - 1))
            a = (scale*((year*ny)**shape))*(Eu**(shape-1))
            b = (shape**-1)*((((year*ny)*Eu)**shape) - 1)
            c = (-scale*(shape**-2))*(((year*ny)*Eu)**shape - 1) + (scale*(shape**-1))*(((year*ny)*Eu)**shape)*np.log((year*ny)*Eu)
            CIyear = (scist.norm.ppf(1-(alpha/2))*((((a**2)*d) + (b*((c*g) + (e*b))) + (c*((b*f) + (c*h))))**0.5))
            CI_z_N_high_year.append(thold + (scale/shape)*(((year*ny*Eu)**shape) - 1) + CIyear)
            CI_z_N_low_year.append(thold + (scale/shape)*(((year*ny*Eu)**shape) - 1) - CIyear)
        
        #Plotting Return Value
        ax.plot(year_array, CI_z_N_high_year, linestyle='--', color='red', alpha = 0.8, lw = 0.9, label = 'Confidence Bands')
        ax.plot(year_array, CI_z_N_low_year, linestyle='--', color='red', alpha = 0.8, lw = 0.9)
        ax.plot(year_array, z_N, color = 'black', label = 'Theoretical Return Level')
        ax.scatter(N, over_thresh, label = 'Empirical Return Level')

        
        text = f'{N_year:.0f} Year Return Level: {x_m:.2f} \u00B1 {CI:.2f}'
        ax.text(.6,.05,text, transform=ax.transAxes)

        ax.set_xscale('log')
        ax.set_xlabel('Return Period')
        ax.set_title('Return Level Plot')
        ax.legend()




"""
Numpy-based functions for estimating tail index
via the Hill Double Bootstrap method.

Process is effectively detailed in Voitalov 2019:
https://journals.aps.org/prresearch/pdf/10.1103/PhysRevResearch.1.033034

See Documentation for more.
"""

from typing import Iterable, Union, Tuple
import numpy as np

import numba as nb
from tqdm.auto import trange

@nb.njit
def A_dani(n1, k1):
    return (np.log(k1)/(2*np.log(n1) - np.log(k1)))**(2*(np.log(n1) - np.log(k1))/(np.log(n1)))

@nb.njit
def A_qi(n1, k1):
    return (1 - (2*(np.log(k1) - np.log(n1))/(np.log(k1))))**(np.log(k1)/np.log(n1) - 1)

@nb.njit
def moments_dbs_prefactor(xi_n, n1, k1):
    """ 
    Function to calculate pre-factor used in moments
    double-bootstrap procedure.
    
    Params
    -------
    xi_n: moments tail index estimate corresponding to
            sqrt(n)-th order statistic.
    n1:   size of the 1st bootstrap in double-bootstrap
            procedure.
    k1:   estimated optimal order statistic based on the 1st
            bootstrap sample.
    
    Returns
    --------
    prefactor: constant used in estimation of the optimal
                stopping order statistic for moments estimator.
    """
    def V_sq(xi_n):
        if xi_n >= 0:
            V = 1. + (xi_n)**2
            return V
        else:
            a = (1.-xi_n)**2
            b = (1-2*xi_n)*(6*((xi_n)**2)-xi_n+1)
            c = (1.-3*xi_n)*(1-4*xi_n)
            V = a*b/c
            return V

    def V_bar_sq(xi_n):
        if xi_n >= 0:
            V = 0.25*(1+(xi_n)**2)
            return V
        else:
            a = 0.25*((1-xi_n)**2)
            b = 1-8*xi_n+48*(xi_n**2)-154*(xi_n**3)
            c = 263*(xi_n**4)-222*(xi_n**5)+72*(xi_n**6)
            d = (1.-2*xi_n)*(1-3*xi_n)*(1-4*xi_n)
            e = (1.-5*xi_n)*(1-6*xi_n)
            V = a*(b+c)/(d*e)
            return V
    
    def b(xi_n, rho):
        if xi_n < rho:
            a1 = (1.-xi_n)*(1-2*xi_n)
            a2 = (1.-rho-xi_n)*(1.-rho-2*xi_n)
            return a1/a2
        elif xi_n >= rho and xi_n < 0:
            return 1./(1-xi_n)
        else:
            b = (xi_n/(rho*(1.-rho))) + (1./((1-rho)**2))
            return b

    def b_bar(xi_n, rho):
        if xi_n < rho:
            a1 = 0.5*(-rho*(1-xi_n)**2)
            a2 = (1.-xi_n-rho)*(1-2*xi_n-rho)*(1-3*xi_n-rho)
            return a1/a2
        elif xi_n >= rho and xi_n < 0:
            a1 = 1-2*xi_n-np.sqrt((1-xi_n)*(1-2.*xi_n))
            a2 = (1.-xi_n)*(1-2*xi_n)
            return a1/a2
        else:
            b = (-1.)*((rho + xi_n*(1-rho))/(2*(1-rho)**3))
            return b

    rho = np.log(k1)/(2*np.log(k1) - 2.*np.log(n1))
    a = (V_sq(xi_n)) * (b_bar(xi_n, rho)**2)
    b = V_bar_sq(xi_n) * (b(xi_n, rho)**2)
    prefactor = (a/b)**(1./(1. - 2*rho))
    return prefactor

@nb.njit
def hill_est_for_alpha(k, y):
    return k / (np.cumsum(np.log(y[:-1])) - k*np.log(y[:-1]))

@nb.njit
def hill_est_for_xi(k,y):
    return np.cumsum(np.log(y[:-1]))/k - np.log(y[1:])

@nb.njit
def second_moment(k, y):
    t1 = np.cumsum(np.log(y[:-1])**2) / k 
    t2 = 2*np.cumsum(np.log(y[:-1]))*np.log(y[1:]) / k
    t3 = np.log(y[1:])**2
    return t1 - t2 + t3

@nb.njit
def third_moment(k,y):
    """
    """
    t1 = (1/k)*np.cumsum(np.log(y[:-1])**3)
    t2 = (3*np.log(y[1:])/k)*np.cumsum(np.log(y[:-1])**2)
    t3 = (3*np.log(y[1:])**2/k)*np.cumsum(np.log(y[:-1]))
    t4 = np.log(y[1:])**3
    M3 = t1 - t2 + t3 - t4
    return M3

@nb.njit
def amse(M1, M2):
    return (M2 - 2*M1**2)**2

@nb.njit
def hill_amse(k,y):
    M1 = hill_est_for_xi(k,y)
    M2 = second_moment(k,y)
    return amse(M1,M2)    

@nb.njit
def moments_amse(k, y):
    M1 = hill_est_for_xi(k, y)
    M2 = second_moment(k,y)    
    M3 = third_moment(k, y)
    xi_2 = M1 + 1 - 0.5*((1 - (M1*M1)/M2))**(-1)
    xi_3 = np.sqrt(0.5*M2) + 1 - (2/3)*(1 / (1 - M1*M2/M3))
    return (xi_2 - xi_3)**2

@nb.njit
def finder_loop(y, n, r, style='hill'):
    amses = np.zeros(n-1)
    for i in range(r):
        sample = np.random.choice(y, n, replace=False)
        sample = np.sort(sample)[::-1]
        k = np.arange(1,n)

        if style == 'hill':
            amses += hill_amse(k,sample)
        elif style =='moments':
            amses += moments_amse(k,sample)

    return amses

def k_finder(y, n, r, kmin, style='hill'):
    kmax = (np.abs(np.linspace(1./n, 1.0, n) - 1)).argmin()
    amses = finder_loop(y, n, r, style)
    amse_for_k = amses / n

    k = np.nanargmin(amse_for_k[kmin:kmax]) + 1 + kmin
    return k

def dbl_bs(y, t=.5, r=500, style='hill', A_type='qi'):
    n = y.size
    n1 = int(np.sqrt(t)*n)
    n2 = int(t*n)
    k = np.arange(1, y.size)
    xi = hill_est_for_xi(k,y)
    xi_n = xi[int(np.floor(n**0.5))-1]

    kmin1, kmin2 = 1,1
    while True:
        k1 = k_finder(y, n1, r, kmin1, style=style)
        k2 = k_finder(y, n2, r, kmin2, style=style)

        if k2 > k1:
            kmin1 += int(0.005*n)
            kmin2 += int(0.005*n)
        else:
            break

    A = A_qi if A_type == 'qi' else A_dani

    if style == 'hill':
        prefactor = A(n1,k1)
    elif style =='moments':
        prefactor = moments_dbs_prefactor(xi_n, n1, k1)
    else:
        raise ValueError('`style` not supported')

    k_star = prefactor*k1**2 / k2
    k_star = np.round(k_star).astype(int)

    if k_star >= n:
        raise ValueError(f'Estimated threshold larger than size of sample data: k {k_star} v. n {n}')
    else:
        k_star = 2 if k_star == 0 else k_star

    return xi[k_star]

def two_tailed_hill_double_bootstrap(
    values:Iterable[float], 
    iters:int=10, 
    return_mean:bool=True,
    pbar_kwargs:dict={},
    ) -> Union[Tuple[Iterable], Tuple[float]]:
    
    left = values[values < values.mean()]
    left = np.sort(-left)[::-1]

    right = values[values > values.mean()]
    right = np.sort(right)[::-1]

    np.seterr(all='ignore')
    shl = np.zeros(iters)
    shr = np.zeros(iters)
    for i in trange(iters, **pbar_kwargs):
        shl[i] = dbl_bs(left, t=.5, r=500, style='hill', A_type='dani')
        shr[i] = dbl_bs(right, t=.5, r=500, style='hill', A_type='dani')
    
    if return_mean:
        return shl.mean(), shr.mean()
    else:
        return shl, shr



"""
Custom functionality for generating ARMA-GARCH time-series forecasts.
"""
from __future__ import annotations

from typing import Iterable, Union, Tuple, TYPE_CHECKING
import warnings
import numpy as np
import pandas as pd
import scipy
import matplotlib.pyplot as plt

import numba as nb

from phat.utils import argsetter, arrayarize, PriceSim
from phat.dists import Phat

import pmdarima
import arch

ALLOWED_DISTS = Union[Phat, scipy.stats._distn_infrastructure.rv_frozen]

###### Numba helper functions to improve GarchForecast speed ######
@nb.njit(parallel=True, cache=True)
def random_normal(
    iters:int=1, 
    periods:int=252
    ) -> np.ndarray:
    """
    Loop creates 2d standard normal residual terms. Numba is much faster
    producing random 2d arrays in this fashion than via the `size` parameter

    Params
    -------
    iters:      int; number of separate simulations to run
    periods:    int; number of periods over which to run simulation

    Returns
    --------
    rvs:        np.ndarray of random draws from standard normal distribtion
    """
    rvs = np.zeros(shape=(iters, periods), dtype=np.float64)
    for i in nb.prange(iters):
        for j in nb.prange(periods):
            rvs[i, j] = np.random.normal()

    return rvs

@nb.njit(cache=True)
def _step_sum(
    params:Iterable[float], 
    values:Iterable[float]
    ) -> float:
    """
    Summation term of an AR(m), MA(n), or GARCH(p) / GARCH(q) process

    Each order-specific, fitted relationship parameter is multiplied by the
    corresponding order-specific lagged value, then summed across all the orders
    for that process.

    Params
    -------
    params:     np.ndarray; fitted parameters corresponding to the process
    values:     np.ndarray; lagged value against which the parameter is multiplied

    """
    sums = np.zeros(values.shape)
    for i in range(values.shape[0]):
        sums[i] = params[i]*values[i]

    return np.sum(sums)    

@nb.njit(cache=True)
def _arma_one_step(
    y:np.ndarray, 
    resids:np.ndarray, 
    constant:float, 
    ar_params:np.ndarray, 
    ma_params:np.ndarray, 
    m:int, 
    n:int
    ) -> float:
    """
    ARMA calculation for single step of forecast. Captures correlation of 
    current period mean return with prior period error terms.

    Params
    -------
    y:              np.ndarray; historical standardized residuals of ARMA-GARCH fit up to `maxlag` periods ago
    resids:         np.ndarray; historical residuals of ARMA-GARCH fit up to `maxlag` periods ago
    constant:       float; constant of the mean process4
    ar_params:      np.ndarray; paramters of the AR process
    ma_params:      np.ndarray; paramters of the MA process
    m:              int; order of the AR process
    n:              int; order of the MA process
    """
    ar = _step_sum(ar_params, y[-m:]) if m != 0 else 0
    ma = _step_sum(ma_params, resids[-n:]) if n != 0 else 0

    return constant + ar + ma

@nb.njit(cache=True)
def _garch_one_step(
    vols:np.ndarray,
    resids:np.ndarray,
    innov:np.ndarray,
    constant:float,
    arch_params:np.ndarray,
    garch_params:np.ndarray,
    p:int, 
    q:int
    ) -> float:
    """
    GARCH calculation for single step of forecast. Captures correlation of
    current period volatility with prior period residuals. The ARCH process is 
    analogous to the MA process for the mean. It is also a subset of GARCH, as:

                                ARCH(q) = GARCH(0,q)

    Params
    -------
    vols:           np.ndarray; historical volatities of ARMA-GARCH fit up to `maxlag` periods ago
    resids:         np.ndarray; historical residuals of ARMA-GARCH fit up to `maxlag` periods ago
    innov:          np.ndarray; random error terms
    constant:       float; constant of the vol process

    arch_params:      np.ndarray; paramters of the ARCH process
    garch_params:      np.ndarray; paramters of the GARCH process
    p:              int; order of the ARCH process
    q:              int; order of the GARCH process
                
    GARCH calculation for single step of forecast. Captures correlation of
    current period volatility with prior volatility. The "generalized" portion of
    "G"ARCH is analogous to the AR process for the mean. Unlike AR, however, it is
    uncommon for a volatility memory to be utilized without residual memory. So,
    you should always pair a `p=1` process with a `q=1` process.

    Remember for GARCH we work with volatilities so we must square them and the residuals.
    """
    arch = _step_sum(arch_params, resids[-q:]**2) if q != 0 else 0
    garch = _step_sum(garch_params, vols[-p:]**2) if p != 0 else 0

    var_t = constant + arch + garch
    resid_t = np.sqrt(var_t)*innov

    return var_t, resid_t
                   
class PROCESSMIXIN:
    """
    Mixin to provide common functionality to MEAN and VOL classes
    """
    @property
    def constant(self) -> float:
        return self.params[0]

    @property
    def max_order(self) -> int:
        return max(self.order)

    @property
    def props(self) -> Union[None, tuple]:
        return self.properties

    def _infer_order(self, params) -> tuple:
        """
        Infers the order of an ARMA or GARCH process.

        If params is even, assume even split between the params. If odd, assumes modulus lag belogs to first order.

        Parameters
        -----------
        params:     iterable; params of process from which order is inferred

        Return
        -------
        order:      tuple; 2x1 tuple with ARMA order (p,q) or GARCH order (m,n)
        """
        if any([params is None, isinstance(params, pd.Series) and params.empty, not params.any()]):
            first_order = second_order = 0
        else:
            first_order = len(params) // 2 + len(params) % 2
            second_order = len(params) // 2
        
        return first_order, second_order

class MEAN(PROCESSMIXIN):
    """
    Containter housing properties of the mean process.

    Typically, ARIMA or constant.

    Parameters
    -----------
    params:     iterable; values of constant plus any m, n process
    order:      tuple; 2x1 of m, n order (m = AR process, n = MA process)
    """
    def __init__(self,
        params:Iterable,
        order:tuple=None,
        ):
        self.params = params.values if hasattr(params, 'values') else params
        self._m, self._n = self._infer_order(params[1:]) if order is None else order
    
    @property
    def m(self) -> int:
        return self._m
    
    @property
    def n(self) -> int:
        return self._n
    
    @property
    def order(self) -> tuple:
        return self.m, self.n
    
    @property
    def ar_params(self) -> Iterable:
        return self.params[1:1+self.m]
    
    @property
    def ma_params(self) -> Iterable:
        return self.params[1+self.m:1+self.m+self.n]
    
    @property
    def properties(self) -> Union[None, tuple]:
        return self.constant, self.ar_params, self.ma_params, self.m, self.n

class VOL(PROCESSMIXIN):
    """
    Containter housing properties of the volatility process.

    Typically, constant, GARCH or some derivative of GARCH.

    Parameters
    -----------
    params:     iterable; values of constant plus any m, n process
    order:      tuple; 2x1 of p, q order
                where:  p = autoregressive to prior volatility 
                        n = autoregressive to prior residuals
    """    
    def __init__(self,
        params:Iterable,
        order:tuple=None
        ):
        self.params = params.values if hasattr(params, 'values') else params
        self._p, self._q = self._infer_order(params[1:]) if order is None else order
    
    @property
    def p(self) -> int:
        return self._p
    
    @property
    def q(self) -> int:
        return self._q
    
    @property
    def order(self) -> tuple:
        return self.p, self.q

    @property
    def arch_params(self):
        return self.params[1:1+self.q]
    
    @property
    def garch_params(self):
        return self.params[1+self.q:1+self.q+self.p]

    @property
    def properties(self) -> Union[None, tuple]:
        return self.constant, self.arch_params, self.garch_params, self.p, self.q

class GarchcastResults:
    """
    Container for outputs from Garchcast.forecast

    `plot` method provides several options for plotting results

    Parameters
    -----------
    values:     np.ndarray; 2d iters x periods array of values from Garchcast projection
    vols:       np.ndarray; 2d iters x periods array of forecast volatilities
    resids:     np.ndarray; 2d iters x periods array of forecast residuals relative to ARMA-GARCH
    """
    def __init__(self, 
        values:np.ndarray, 
        vols:np.ndarray, 
        resids:np.ndarray
        ):
        self.values = values
        self.vols = vols
        self.resids = resids

        self.iters, self.periods = values.shape

    @property
    def rets(self) -> np.ndarray:
        return 1 + self.values/100

    @property
    def vars(self) -> np.ndarray:
        return self.vols**2

    @property
    def sim(self) -> np.ndarray:
        if hasattr(self, '_sim'):
            return self._sim
        else:
            raise ValueError('You must plot a price simulation or call `_make_sim` directly')

    @argsetter(['iters', 'periods'])
    def _make_sim(self, iters:int, periods:int, p:float=1):
        """
        Creates a price simulator used to generate images
        """
        self._sim = PriceSim(p0=p, periods=periods, n=iters)
        return self._sim

    def plot(
        self, 
        kind:str='vol', 
        ax:Union[plt.Axes, Iterable[plt.Axes]]=None, 
        *args, **kwargs
        ) -> Union[plt.Axes, Iterable[plt.Axes]]:
        """
        Main method used to generate plots

        Params
        -------
        kind:    str; kind of plot. Accepts 'vol', 'var', 'price', or 'end_price'
        ax:      matplotlib Axes object(s); if one is not provided, it will be created.

        Return
        -------
        ax:      matplotlib Axes object(s)
        """
        multi_ax = ['price']
        if ax is None and kind not in multi_ax:
            fig, ax = plt.subplots(1,1,figsize=(10,6))

        if kind == 'vol':
            ax = self._plot_vol(ax, *args, **kwargs)
        elif kind == 'var':
            ax = self._plot_var(ax, *args, **kwargs)
        elif kind == 'price':
            if len(args) > 1:
                p = args[0]
            if len(args) > 2:
                n = args[2]
            if 'p' in kwargs:
                p = kwargs['p']
                kwargs.pop('p')
            if 'n' in kwargs:
                n = kwargs['n']
                kwargs.pop('n')
            
            ax = self._plot_price(p=p, n=n, axes=ax, *args, **kwargs)
        elif kind == 'end_price':
            if len(args) > 1:
                p = args[0]
            if 'p' in kwargs:
                p = kwargs['p']
                kwargs.pop('p')

            return self._plot_end_price(p=p, ax=ax, *args, **kwargs)
        else:
            raise ValueError(f'`kind:` {kind} not supported')

        return ax

    def _plot_vol(self, ax:plt.Axes, *args, **kwargs) -> Union[plt.Axes, Iterable[plt.Axes]]:
        ax.plot((self.vols).mean(axis=0))
        ax.set_xlabel('Days')
        plt.suptitle('Phat-GARCH Forecast: Conditional Volatility')

        return ax
        
    def _plot_var(self, ax, *args, **kwargs) -> Union[plt.Axes, Iterable[plt.Axes]]:
        ax.plot((self.vols**2).mean(axis=0), *args, **kwargs)
        ax.set_xlabel('Days')
        plt.suptitle('Phat-GARCH Forecast: Conditional Variance')
        
        return ax

    def _plot_price(self, 
        p:float, 
        n:int, 
        axes:Iterable[plt.Axes], 
        *args, **kwargs
        ) -> Iterable[plt.Axes]:
        sim = self._make_sim(p=p)
        
        all_axes = []
        for i in np.random.randint(0, self.iters, size=n):
            _, __, chart_axes = sim.sim(rets=(1 + self.values[i]/100), axes=axes, show_chart=True, *args, **kwargs)
            all_axes.append(chart_axes)
        
        if axes is None:
            return np.array(all_axes)
        else:
            return axes

    def _plot_end_price(self, 
        p:float, 
        ax:plt.Axes, 
        *args, **kwargs
        ) -> Tuple[plt.Axes, np.ndarray, np.ndarray]:
        sim = self._make_sim(p=p)
        _, P, (ax, bins) = sim.sims(rets=1 + self.values/100, ax=ax, show_chart=True, *args, **kwargs)
        return ax, P, bins

class Garchcaster:
    """
Class for forecasting time series via ARMA-GARCH process.

Instances of a `pmdarima`-based ARIMA object and/or an ARCHModelResult from the 
`arch` package can be provided directly. Alternatively, the user can provide all
residuals, standardized residuals, conditional volatilities, the parameters for 
the ARMA-GARCH process, and the order.

Params:
    arma: instance of pmdarima ARIMA class OR iterable of floats 
        Either provide a pmdarima instance or the params derived from any arima process.
        If None, will search for AR process in `garch`, otherwise assumes ConstantMean.
    garch: instance of arch ARCHModelResult class OR iterable of floats
        either provide an ARCHModelResult instance or the params derived from any garch process
    y:  float or iterable of floats 
        iters x periods standardized residuals after filtering an ARMA-GARCH. Only required if
        garch is not ARCHModelResult type.
    vols: float or iterable of floats
        iters x periods conditional volatilities resulting from an ARMA-GARCH process. Only required if 
        garch is not ARCHModelResult type.
    resids: float or iterable of floats
        iters x periods residuals after filtering an ARMA-GARCH. Only required if garch is not ARCHModelResult type.
    iters: int 
        number of iterations of forecast
    periods: int 
        number of periods in each iteration of the forecast
    dist: Phat or scipy rv_frozen class 
        distribution used to generate residuals
    use_backcast: bool 
        use procedure to smooth values. See `here <https://arch.readthedocs.io/en/latest/univariate/generated/generated/arch.univariate.GARCH.backcast.html>`_.
    """
    GARCH_HAS_MEAN_PROCESS = False

    def __init__(self,
        garch:Union[arch.univariate.base.ARCHModelResult, Iterable[float]]=None,
        arma:Union[pmdarima.arima.arima.ARIMA, Iterable[float]]=None,
        y:Iterable[float]=None,
        vols:Iterable[float]=None,
        resids:Iterable[float]=None,
        iters:int=1, 
        periods:int=252,
        order:tuple=None,
        dist:ALLOWED_DISTS=None,
        use_backcast:bool=True,
        ):
        ### First, set self.mean
        if isinstance(arma, pmdarima.arima.arima.ARIMA):
            m, _, n = arma.order
            mean_order = (m,n)
            mean_params = arma.params()[:5]
        elif isinstance(garch, arch.univariate.base.ARCHModelResult):
            if isinstance(garch.model, arch.univariate.mean.ConstantMean):
                mean_order = (0, 0)
                mean_params = garch.params[:1]
            else:
                mean_order = (garch.model.lags[0].size, 0)
                mean_params = garch.params[:mean_order[0]+1]
                self.GARCH_HAS_MEAN_PROCESS = True
        elif arma is not None:
            mean_order = list(order)[:2] if order is not None else None
            mean_params = arrayarize(arma)
        
        self.mean = MEAN(mean_params, order=mean_order)

        ### Second, set self.vol
        if isinstance(garch, arch.univariate.base.ARCHModelResult):
            if hasattr(garch.model.volatility, 'p'):
                vol_order = (garch.model.volatility.p, garch.model.volatility.q)
            else:
                vol_order = (0, 0)

            vol_params = garch.params[mean_order[0]+1:] if self.GARCH_HAS_MEAN_PROCESS else garch.params[1:]
        elif isinstance(garch, (list, tuple, set, np.ndarray)):
            vol_params = arrayarize(garch)
            vol_order = list(order)[2:] if order is not None else None
        else:
            vol_order = (0, 0)

        self.vol = VOL(vol_params, order=vol_order)
        self.order = (*self.mean.order, *self.vol.order)

        if order is not None and order != self.order:
            msg = 'The `order` parameter provided does not match the order'
            msg += ' inferred from the models provided and has been OVERRIDDEN. The `order` parameter'
            msg += ' is not required when passing `arch` and/or pmdarima models to Garchcaster.'
            warnings.warn(msg)

        self.maxlag = max(self.order)

        ### Third, assign values
        if isinstance(garch, arch.univariate.base.ARCHModelResult):
            if any([y is not None, vols is not None, resids is not None]):
                raise ValueError('If `garch` is ARCHModelResult do not provide `y`, `vol`, or `resids`')
            else:
                y = garch.std_resid.values[self.maxlag:].copy()
                vols = garch.conditional_volatility.values[self.maxlag:].copy()
                resids = garch.resid.values[self.maxlag:].copy()
        else:
            y = arrayarize(y).copy()
            vols = arrayarize(vols).copy()
            resids = arrayarize(resids).copy()
        
        self.periods = periods
        self.iters = iters
   
        self.y_hist = y[-self.maxlag:]
        self.vols_hist = vols[-self.maxlag:]
        self.resids_hist = self._backcast(resids, self.maxlag) if use_backcast else resids[:-self.maxlag:]
        
        self._dist = dist
    
    @property
    def dist(self) -> ALLOWED_DISTS:
        if self._dist is None:
            NotImplementedError('You must pass `dist` parameter')
        else:
            return self._dist

    @argsetter(['iters', 'periods'], flat=True)
    def _sample_innovs(self, 
        iters:int=None, 
        periods:int=None, 
        dist:ALLOWED_DISTS=None,
        seed:int=None,
        ) -> np.ndarray:
        """
        Generates innovations according to the provided distribution. If a distribution is not
        provided, default is Gaussian.

        Params
        -------
        iters:          int; number of iterations of forecast
        periods:        int; number of periods in each iteration of the forecast
        dist:           Phat or scipy rv_frozen class; distribution used to generate residuals
        seed:           int; random state for innovation determination if `innovs` is not provided


        Return
        -------
        innovs:         np.ndarray
        """
        if self._dist is None and dist is None:
            innovs = random_normal(iters, periods)
        else:
            dist = dist if dist is not None else self._dist
            rvfunc = dist.std_rvs if hasattr(dist, 'std_rvs') else dist.rvs
            innovs = rvfunc(size=(iters, periods), seed=seed)

        return innovs

    def _backcast(self, resids, maxlag):
        power = 2
        tau = min(75, resids.shape[0])
        w = 0.94 ** np.arange(tau)
        w = w / sum(w)
        backcast = np.sum((abs(resids[:tau]) ** power) * w)

        return np.repeat(backcast, maxlag)

    @staticmethod
    @nb.njit(cache=True)
    def _forecast_looper(
        y:np.ndarray,
        vols:np.ndarray,
        resids:np.ndarray,
        innovs:np.ndarray,
        iters:int, 
        periods:int,
        mean_args:tuple,
        vol_args:tuple,
        ) -> Tuple[np.ndarray]:
        """
        Main iterative, feed-forward function for forecasting an ARMA-GARCH process.
        
        Process
        --------
        To forecast a single-step period within a single iteration of the forecast

        1. First find the current variance. This requires historical variances for the
        p-requested lags.
        2. Find the residual term, which is a function of the volatility and the
        error term provided (or calculated as Standard Normal)
        3. Find mean value from ARMA
        4. Reassign y, errs, vols to feed forward into next iteration

        Params
        -------
        y:              np.ndarray; historical standardized residuals of ARMA-GARCH fit up to `maxlag` periods ago
        vols:           np.ndarray; historical volatilities of ARMA-GARCH fit up to `maxlag` periods ago
        resids:         np.ndarray; historical residuals of ARMA-GARCH fit up to `maxlag` periods ago
        innovs:         iters x periods np.ndarray of garch residuals
        iters:          int; number of iterations of forecast
        periods:        int; number of periods in each iteration of the forecast
        dist:           Phat or scipy rv_frozen class; distribution used to generate residuals
        mean_args:      tuple; properties of the mean process use in `_arma_one_step`    
        vol_args:       tuple; properties of the vol process use in `_garch_one_step`
        """
        y_ = np.zeros((iters, periods))
        vols_ = np.zeros((iters, periods))
        resids_ = np.zeros((iters, periods))
        for i in range(iters):
            y_i = y.copy()
            vols_i = vols.copy()
            resids_i = resids.copy()
            for t in range(periods):
                mu_t = _arma_one_step(y_i, resids_i, *mean_args)
                var_t, resid_t = _garch_one_step(vols_i, resids_i, innovs[i, t], *vol_args)

                y_[i, t] = mu_t + resid_t
                vols_[i, t] = np.sqrt(var_t)
                resids_[i, t] = resid_t
                
                y_i[-1] = y_[i, t]
                vols_i[-1] = vols_[i, t]
                resids_i[-1] = resids_[i, t]                
            
                if y.size > 1:
                    y_i[-y_i.size:-1] = y_i[-y_i.size+1:]
                    vols_i[-y_i.size:-1] = vols_i[-y_i.size+1:]
                    resids_i[-y_i.size:-1] = resids_i[-y_i.size+1:]

                # if np.isnan(y_).any() or np.isnan(y).any():
                #     print (i, t, mu_t, resid_t, innovs[i,t], y, vols, resids)
        return y_, vols_, resids_

    @argsetter(['dist', 'iters', 'periods'], flat=True)
    def forecast(self, 
        iters:int=None, 
        periods:int=None, 
        dist:ALLOWED_DISTS=None,
        innovs:Iterable=None,
        seed:int=None,
        ) -> GarchcastResults:
        """
        User-facing interface with forecast generator

        Params:
            iters: int
                number of iterations of forecast
            periods: int
                number of periods in each iteration of the forecast
            dist: Phat or scipy rv_frozen class
                distribution used to generate residuals
            innovs: np.ndarray
                iters x periods of garch residuals
            seed: int
                random state for innovation determination if `innovs` is not provided
        """
        if innovs is None:
            innovs = self._sample_innovs(iters, periods, dist=dist, seed=seed)

        results = Garchcaster._forecast_looper(
            self.y_hist, 
            self.vols_hist, 
            self.resids_hist, 
            innovs,
            iters, periods,
            self.mean.props,
            self.vol.props
        )
        return GarchcastResults(*results)
